"""
Mesh Info stores the mesh requirements.
"""
import json
import numpy
from mesh_generator.ui.mesh_box import MeshBox


class MeshInfo:
    """
    UI Mesh dictionary. The dictionary will include the mesh requirements inputted by the user in
    tkinter ui - tkinter_mainapp.py.
    -lower and upper bounds for theta/phi/radial coordinates.
    - BG_RATIO - the ratio in regions that the user does not specify a fixed ds. (regions you do not care about) t/p
    - FG_RATIO - the maximum ration in regions that the user does specify a fixed ds. (regions you do care about) t/p
    - box_list - a list of 3d boxes with mesh segment requirements.
    """

    def __init__(self, lower_bnd_t=0., upper_bnd_t=numpy.pi, lower_bnd_p=0, upper_bnd_p=2 * numpy.pi, lower_bnd_r=1.0,
                 upper_bnd_r=2.5, BG_RATIO=1.06, FG_RATIO=1.03, RADIAL_BG_RATIO=1.06, RADIAL_FG_RATIO=1.03,
                 box_list=numpy.array([], dtype=MeshBox)):
        self.lower_bnd_t = float(lower_bnd_t)
        self.upper_bnd_t = float(upper_bnd_t)
        self.lower_bnd_p = float(lower_bnd_p)
        self.upper_bnd_p = float(upper_bnd_p)
        self.lower_bnd_r = float(lower_bnd_r)
        self.upper_bnd_r = float(upper_bnd_r)
        self.BG_RATIO = float(BG_RATIO)
        self.FG_RATIO = float(FG_RATIO)
        self.RADIAL_BG_RATIO = float(RADIAL_BG_RATIO)
        self.RADIAL_FG_RATIO = float(RADIAL_FG_RATIO)
        self.box_list = box_list

    def __str__(self):
        return json.dumps(
            self.json_dict(), indent=4, default=lambda o: o.json_dict())

    def json_dict(self):
        return {
            'lower_bnd_t': self.lower_bnd_t,
            'upper_bnd_t': self.upper_bnd_t,
            'lower_bnd_p': self.lower_bnd_p,
            'upper_bnd_p': self.upper_bnd_p,
            'lower_bnd_r': self.lower_bnd_r,
            'upper_bnd_r': self.upper_bnd_r,
            'BG_RATIO': self.BG_RATIO,
            'FG_RATIO': self.FG_RATIO,
            'RADIAL_BG_RATIO': self.RADIAL_BG_RATIO,
            'RADIAL_FG_RATIO': self.RADIAL_FG_RATIO,
            'box_list': self.type_box_list()
        }

    def insert_mesh_box(self, index, new_cube):
        """
            Insert a 3d Box to the list of UI boxes.
        """
        self.box_list = numpy.insert(self.box_list, index, new_cube)

    def append_mesh_box(self, new_cube):
        """
            Append a 3d box to the list of UI boxes.
        """
        self.box_list = numpy.append(self.box_list, new_cube)

    def type_box_list(self):
        if isinstance(self.box_list, list):
            return self.box_list
        else:
            return self.box_list.tolist()



"""
Plotting the psi mesh data results saved in the file mesh_res[t/p/r].dat
"""
import numpy as np
from matplotlib import pyplot as plt
import time
import os


def plot_mesh_res(adjusted_mesh: dict, save_plot: bool, mesh_res_file_name: str, show_plot: bool, label: str,
                  save_plot_path: str, plot_file_name: str):
    """
       plot the mesh points in tmp_mesh_r.dat

       Arguments:
           :param mesh_res_file_name: name of file with mesh res. Ex) mesh_res_r.dat
           :param plot_file_name: name of the png file saved.
           :param save_plot_path: path to save png plot.
           :param show_plot: bool: if true it will show the matplotlib figure. Interactive widget.
           :param save_plot: dictionary with adjusted mesh segments.
           :param adjusted_mesh: bool: if true it will save the plot as a png in plot file.
           :param label:  't'/'p'/'r'
       """
    xvals, yvals, rvals = [], [], []

    with open(str(mesh_res_file_name), "r") as f:
        fig, axs = plt.subplots(2, figsize=(7, 6.5))
        for idx, line in enumerate(f):
            if idx >= 1:
                for word in line.split()[1:2]:
                    xvals.append(float(word))
                for word in line.split()[2:3]:
                    yvals.append(float(word))
                for word in line.split()[3:4]:
                    rvals.append(float(word))

    # Set the plot title by the mesh type (theta/phi/radial)
    if label == 't':
        axs[0].set_xlabel(' \u03B8 ')
        axs[1].set_xlabel(' \u03B8 ')
        axs[0].set_ylabel('\u0394 \u03B8')
        title = "Theta Mesh Spacing"
    if label == 'p':
        axs[0].set_xlabel(' \u03C6 ')
        axs[1].set_xlabel(' \u03C6 ')
        axs[0].set_ylabel('\u0394 \u03C6')
        plt.axvline(x=0, dashes=[6, 2], color='gray')
        plt.axvline(x=np.pi * 2, dashes=[6, 2], color='gray')
        title = "Phi Mesh Spacing"
    if label == 'r':
        axs[0].set_xlabel(' r ')
        axs[1].set_xlabel(' r ')
        axs[0].set_ylabel('\u0394 r')
        title = "Radial Mesh Spacing"

    # total number of points is also the total number of rows in bin/mesh_res.txt file.
    num_of_points = str(idx)
    axs[0].set_title(title + ", n = " + num_of_points, y=1.3)  # add it to the title
    # plot the legacy mesh
    axs[0].plot(xvals, yvals, label="Legacy mesh")
    # plot the mesh ratio
    axs[1].plot(xvals, rvals, label='Ratio')
    axs[1].set_title("Ratio", y=1.3)
    seg_label = "Input segment requirements"
    adj_label = "Adjusted segments"
    # plot step 2 of creating a mesh. The segments after calling the function resolve_mesh_segments() in mesh src/
    for i in range(0, 2):
        for segments in adjusted_mesh['segment_list']:
            s1 = segments['s1'] + (adjusted_mesh['phi_shift'])
            s0 = segments['s0'] + (adjusted_mesh['phi_shift'])
            axs[0].hlines(max(segments['ds0'], segments['ds1']), s0, s1, linewidth=3, label=seg_label)
            seg_label = "_nolegend_"
            axs[i].axvline(x=s1, dashes=[6, 2], color='gray')
            axs[i].axvline(x=s0, dashes=[6, 2], color='gray')
            axs[0].plot([s1, s0], [segments['ds1'], segments['ds0']], linewidth=1, color='aqua', label=adj_label)
            adj_label = "_nolegend_"
        axs[i].axvline(x=adjusted_mesh['upper_bnd'], dashes=[6, 2], color='gray')
        axs[i].axvline(x=adjusted_mesh['lower_bnd'], dashes=[6, 2], color='gray')

    fig.canvas.set_window_title(title)
    axs[0].legend(bbox_to_anchor=(0, 1.02, 1, 0.2), loc="lower left", mode="expand", ncol=3)
    axs[1].legend(bbox_to_anchor=(0, 1.02, 1, 0.2), loc="lower left", mode="expand", ncol=1)
    plt.tight_layout()

    if save_plot:
        if plot_file_name is None:
            plot_file_name = str(time.ctime()) + " " + title + ".png"

        plot_path = os.path.join(save_plot_path, plot_file_name)
        plt.savefig(plot_path)

    if show_plot:
        plt.show()





"""Mesh Generator PSI

Create a 1D Mesh in Python.

Important subpackages:
src- mesh calculations.
bin- output .dat files, optimization, plotting.


Documentation: https://q.predsci.com/docs/mesh_generator/

"""
from mesh_generator import src
from mesh_generator import bin
from mesh_generator import hdf
from mesh_generator.src.mesh import Mesh
from mesh_generator.src.mesh_segment import MeshSegment
from mesh_generator.bin.call_psi_mesh_tool import create_psi_mesh
from mesh_generator.hdf.read_mesh_res import read_tmp_file
from mesh_generator.hdf.write_1d_hdf_files import write_1d_mesh_hdf
from mesh_generator.ui.upload_magnetogram import MeshGeneratorUI



""" Mesh Box stores the mesh segments requirements in a 3d box format. """
import json
import numpy


class MeshBox:
    """
    UI Mesh box. Rectangles interactively drawn by the user on the magnetogram plot found in the main UI
    : (tkinter_mainapp.py). These are 3d boxes in spherical coordinates.
    t - theta, p - phi, r - radial.
    t0 - coordinate of begin segment. t1- coordinate of end segment (theta).
    ds- the resolution in theta/phi segments.
    tp_var_ds - the ratio growth in theta/phi segments.
    radial_ds- resolution in the radial segments.
    radial_var_ds - the ratio growth in radial segments.
    index - the order of the mesh box in the box_list.
    """
    def __init__(self, t0, t1, p0, p1, r0=None, r1=None, ds=None, radial_ds=None, radial_var_ds=None, tp_var_ds=None):
        self.t0 = numpy.float64(t0)
        self.t1 = numpy.float64(t1)
        self.p0 = numpy.float64(p0)
        self.p1 = numpy.float64(p1)
        self.r0 = r0
        self.r1 = r1
        self.ds = numpy.float64(ds)
        self.radial_ds = radial_ds
        self.radial_var_ds = radial_var_ds
        self.tp_var_ds = tp_var_ds

    def __str__(self):
        return json.dumps(
            self.json_dict(), indent=2, default=lambda o: o.json_dict())

    def json_dict(self):
        return {
            't0': self.t0,
            't1': self.t1,
            'p0': self.p0,
            'p1': self.p1,
            'r0': self.r0,
            'r1': self.r1,
            'ds': self.ds,
            'radial_ds': self.radial_ds,
            'radial_var_ds': self.radial_var_ds,
            'tp_var_ds': self.tp_var_ds
        }



""" This is the first window of the mesh UI. Here the user will upload magnetogram, upload existing json with previous
UI session info. """
import tkinter as tk
import tkinter.filedialog
import os
import json
from mesh_generator.ui.auto_mesh_from_dict import MeshRes
from mesh_generator.ui.tkinter_mainapp import MeshApp
from mesh_generator.ui.mesh_info import MeshInfo
import numpy as np


class MeshGeneratorUI:
    def __init__(self):
        # main window.
        self.root = tk.Tk()
        self.root.wm_title("Upload Magnetogram")  # title of the window.
        self.root.config(background="gray70")  # background color of the window.
        self.root.geometry("815x210")  # window size.
        self.root.minsize(width=810, height=210)  # minimum window size that the user can adjust.
        self.root.maxsize(width=820, height=215)  # maximum window size that the user can adjust.
        self.Frame = tk.Frame(self.root, width=810, height=210, relief=tk.SUNKEN, background="gray92")
        self.Frame.grid(row=0, column=0, sticky=tk.NSEW)  # position of the frame on the main window

        # hdf file path that will be passed to the MainApp, there the magnetogram will be plotted using matplotlib.
        self.hdf_file_path = None
        # json file containing previous session mesh requirements.
        self.json_file_path = None
        # folder where mesh results will be saved:
        self.save_folder_path = None

        # print instructions on window.
        tk.Label(self.Frame, text="Step 1: Upload Magnetogram 2D hdf file, input path:").grid(row=0, column=0,
                                                                                              sticky=tk.W, columnspan=2)

        tk.Label(self.Frame, text="Step 2: (Optional) Upload a json file constaining previous session mesh "
                                  "requirements, otherwise will start a new UI session. "). \
            grid(row=2, column=0, sticky=tk.W, columnspan=6, rowspan=2)

        tk.Label(self.Frame, text="Step 3: Provide a folder path where mesh results will be saved. "). \
            grid(row=6, column=0, sticky=tk.W, columnspan=6, rowspan=2)

        tk.Label(self.Frame, text="Step 4: If json file is provided then you can skip interactive UI main window and "
                                  "directly compute mesh results. "). \
            grid(row=10, column=0, sticky=tk.W, columnspan=6, rowspan=2)

        # create a textbox for user to input hdf file path.
        self.upload_mag_path = tk.Entry(self.Frame, width=70)
        self.upload_mag_path.insert(tk.END, '')
        self.upload_mag_path.grid(row=1, column=0, columnspan=5, sticky=tk.W)
        tk.Button(self.Frame, text='browse', command=self.find_path_hdf).grid(row=1, column=4, sticky=tk.W)

        # create a textbox for user to input json file path.
        self.upload_json_path = tk.Entry(self.Frame, width=70)
        self.upload_json_path.insert(tk.END, '')
        self.upload_json_path.grid(row=5, column=0, columnspan=4, sticky=tk.W)
        tk.Button(self.Frame, text='browse', command=self.find_path_json).grid(row=5, column=4, sticky=tk.W)

        # create a textbox for user to input dir path where files will be saved.
        self.save_dir = tk.Entry(self.Frame, width=70)
        self.save_dir.insert(tk.END, '')
        self.save_dir.grid(row=8, column=0, columnspan=4, sticky=tk.W)
        tk.Button(self.Frame, text='browse', command=self.find_dir).grid(row=8, column=4, sticky=tk.W)

        # skip interactive and open main ui window.
        tk.Button(self.Frame, text="Skip interactive", command=self.skip_interactive). \
            grid(row=12, column=1, sticky=tk.NSEW)
        tk.Button(self.Frame, text="Mesh UI main window", command=self.main_app_window_activate). \
            grid(row=12, column=2, sticky=tk.NSEW)

        self.root.mainloop()

    def find_path_hdf(self):
        """ hdf file browser function. """
        if self.check_hdf_path_is_valid():
            self.hdf_file_path = tk.filedialog.askopenfilename(parent=self.Frame, title='Select a File',
                                                               initialdir=os.path.dirname(self.hdf_file_path),
                                                               filetypes=[('hdf file', '.hdf')])
        else:
            self.hdf_file_path = tk.filedialog.askopenfilename(parent=self.Frame, title='Select a File', initialdir="/",
                                                               filetypes=[('hdf file', '.hdf')])
        self.upload_mag_path.delete("0", "end")
        self.upload_mag_path.insert(tk.END, self.hdf_file_path)

    def find_path_json(self):
        """ json file browser function. """
        if self.check_json_path_is_valid():
            self.json_file_path = tk.filedialog.askopenfilename(parent=self.Frame, title='Select a File',
                                                                initialdir=os.path.dirname(self.json_file_path),
                                                                filetypes=[('json file', '.json')])
        else:
            self.json_file_path = tk.filedialog.askopenfilename(parent=self.Frame, title='Select a File',
                                                                initialdir="/", filetypes=[('json file', '.json')])
        self.upload_json_path.delete("0", "end")
        self.upload_json_path.insert(tk.END, self.json_file_path)

    def find_dir(self):
        """ json file browser function. """
        if self.check_save_dir_path_is_valid():
            self.save_folder_path = tk.filedialog.askdirectory(parent=self.Frame, title='Select a Folder',
                                                               initialdir=self.save_folder_path)
        else:
            self.save_folder_path = tk.filedialog.askdirectory(parent=self.Frame, title='Select a Folder',
                                                               initialdir="/")
        self.save_dir.delete("0", "end")
        self.save_dir.insert(tk.END, self.save_folder_path)

    def get_save_dir_path(self):
        """ return the latest save dir path. """
        self.save_folder_path = self.save_dir.get()

    def get_hdf_path(self):
        """ return the latest hdf file path. """
        self.hdf_file_path = self.upload_mag_path.get()

    def get_json_path(self):
        """ return the latest json file path. """
        self.json_file_path = self.upload_json_path.get()

    def check_hdf_path_is_valid(self):
        """ check if user hdf path input is valid."""
        try:
            self.get_hdf_path()
            return os.path.exists(path=self.hdf_file_path)
        except:
            return False

    def check_json_path_is_valid(self):
        """ check if user json path input is valid."""
        try:
            self.get_json_path()
            return os.path.exists(path=self.json_file_path)
        except:
            return False

    def check_save_dir_path_is_valid(self):
        """ check if user save dir path input is valid."""
        try:
            self.get_save_dir_path()
            return os.path.exists(path=self.save_folder_path)
        except:
            try:
                print("Save folder path is not a current directory, hence creating this folder.")
                os.makedirs(path=self.save_folder_path, exist_ok=True)
                return True
            except:
                return False

    def skip_interactive(self):
        """ skip interative button function,
        this will call a save in directory window and will do all the calculation in the backend.
        """
        # make sure the json file provided is valid
        if self.check_json_path_is_valid() is False:
            print("The json file path provided is invalid. Use the browser button to find the exact location of the "
                  "previous session json file.")
            return 0

        # make sure the json file provided is valid
        if self.check_save_dir_path_is_valid() is False:
            print("The save folder path provided is invalid. Use the browser button to find the exact location of the "
                  "folder where mesh results will be saved.")
            return 0

        elif self.check_json_path_is_valid() and self.check_save_dir_path_is_valid():
            # calculate the mesh given the requirements in the json file.

            MeshRes(save_mesh_dir=self.save_folder_path, json_file_path=self.json_file_path)
            print("Mesh files are written successfully.")

    def main_app_window_activate(self):
        """ activate the main app window. """
        if self.check_hdf_path_is_valid():
            if self.check_save_dir_path_is_valid():
                if self.check_json_path_is_valid():
                    self.open_main_window_with_existing_req()
                else:
                    self.open_main_window_without_existing_req()
            else:
                print("save folder path is not valid. Please re-enter the folder path.",
                      self.save_folder_path)
        else:
            print("hdf file is not valid. Please re-enter the hdf file path.",
                  self.hdf_file_path)

    def open_main_window_with_existing_req(self):
        """ if hdf file and json file are valid then open the main window with existing mesh requirements. """
        with open(self.json_file_path, 'r') as data:
            ui_dict = json.load(data)
            MeshApp(hdf_file_path=self.hdf_file_path, save_mesh_file_path=self.save_folder_path,
                    ui_session_data=ui_dict, UseExistingInput=True)

    def open_main_window_without_existing_req(self):
        """Generate the GUI without any previous mesh requirements. """
        my_ui_session = MeshInfo(lower_bnd_t=0, upper_bnd_t=np.pi, lower_bnd_p=0, upper_bnd_p=2 * np.pi,
                                 lower_bnd_r=1.0, upper_bnd_r=2.5)
        MeshApp(hdf_file_path=self.hdf_file_path, save_mesh_file_path=self.save_folder_path,
                ui_session_data=my_ui_session, UseExistingInput=False)


if __name__ == "__main__":
    MeshGeneratorUI()



"""
Template to write mesh_header.dat file that will later be used for
MAS MHD code.
"""

import textwrap


def write_mesh_header(tmp_mas_p_file: str, tmp_mas_t_file: str, tmp_mas_r_file: str, mesh_header_file_path: str):
    """ write mesh_header.dat file in bin folder. This formatting is copied from Cooper's IDL code.
    :param mesh_header_file_path: path and file name of resulted mesh_header.dat file.
    :param tmp_mas_r_file: mas radial mesh file path.
    :param tmp_mas_t_file: mas theta mesh file path.
    :param tmp_mas_p_file: mas phi mesh file path.
    """

    # open mas theta mesh file and read the input.
    with open(tmp_mas_t_file, "r") as theta:
        data_t = theta.readlines()

    # open mas phi mesh file and read the input.
    with open(tmp_mas_p_file, "r") as phi:
        data_p = phi.readlines()

    # open mas radial mesh file and read the input.
    with open(tmp_mas_r_file, "r") as radial:
        data_r = radial.readlines()

    # open mesh header file.
    with open(mesh_header_file_path, "w") as mesh_header:
        mesh_header.write(' &topology\n')

        # nr = total number of points. 2nd line in tmp_r file.
        mesh_header.write(data_r[1])

        # nt = total number of points. 1st line in tmp_t file.
        mesh_header.write(data_t[0])

        # np = total number of phi points. 1st line in tmp_p file.
        mesh_header.write(data_p[0])

        mesh_header.write(' /\n')
        mesh_header.write(' &data\n')

        # r mesh info.
        mesh_header.write('  r0=1.00\n')
        mesh_header.write(data_r[0])
        mesh_header.write(textwrap.fill(data_r[2], width=70, subsequent_indent="    ") + "\n")
        mesh_header.write(textwrap.fill(data_r[3], width=70, subsequent_indent="    ") + "\n")
        mesh_header.write(textwrap.fill(data_r[4], width=70, subsequent_indent="    ") + "\n")

        # theta mesh info.
        mesh_header.write(textwrap.fill(data_t[1], width=70, subsequent_indent="    ") + "\n")
        mesh_header.write(textwrap.fill(data_t[2], width=70, subsequent_indent="    ") + "\n")
        mesh_header.write(textwrap.fill(data_t[3], width=70, subsequent_indent="    ") + "\n")

        # phi mesh info.
        mesh_header.write(textwrap.fill(data_p[1], width=70, subsequent_indent="    ") + "\n")
        mesh_header.write(textwrap.fill(data_p[2], width=70, subsequent_indent="    ") + "\n")
        mesh_header.write(textwrap.fill(data_p[3], width=70, subsequent_indent="    ") + "\n")
        mesh_header.write(textwrap.fill(data_p[4], width=70, subsequent_indent="    ") + "\n")


if __name__ == "__main__":
    from mesh_generator.bin.tmp_mas_template import write_tmp_mas_file

    write_tmp_mas_file(mesh_type="t", tmp_mesh_dir_name="tmp_mesh_t.dat", tmp_mas_dir_name="tmp_mas_t.dat")
    write_tmp_mas_file(mesh_type="p", tmp_mesh_dir_name="tmp_mesh_p.dat", tmp_mas_dir_name="tmp_mas_p.dat")
    write_tmp_mas_file(mesh_type="r", tmp_mesh_dir_name="tmp_mesh_r.dat", tmp_mas_dir_name="tmp_mas_r.dat")
    write_mesh_header(tmp_mas_p_file="tmp_mas_p.dat",
                      tmp_mas_t_file="tmp_mas_t.dat",
                      tmp_mas_r_file="tmp_mas_r.dat",
                      mesh_header_file_path="mesh_header.dat")




"""

Usage:
1. Call psi_tools fortran mesh to produce .dat output files with mesh points.
2. Optimize legacy total number of mesh points.
3. plot solution.

For more details: https://q.predsci.com/docs/mesh_generator/

Important modules:
call_fortran_mesh.py
plot_mesh_res.py
"""

from mesh_generator.bin.call_psi_mesh_tool import create_psi_mesh
from mesh_generator.bin.mesh_header_template import write_mesh_header
from mesh_generator.bin.shell_command import run_shell_command
from mesh_generator.bin.plot_mesh_res import plot_mesh_res
from mesh_generator.bin.tmp_mas_template import write_tmp_mas_file




""" In this Python module, we write 1D hdf files with r/t/p mesh spacing. """
from mesh_generator.hdf.psihdf import wrhdf_1d
import numpy as np
from scipy.interpolate import interp1d
import os


def write_1d_mesh_hdf(outdir: str, mesh_name: str, r, t, p, fac=1.0):
    """ Interpolate and build half meshes.
    :param outdir: directory to save mesh file.
    :param mesh_name: name of mesh file.
    :param fac: factor for interpolation spacing.
    :param p: phi mesh spacing, array.
    :param t: theta mesh spacing, array.
    :param r: radial mesh spacing, array.
    """
    # number of elements in r/p/t arrays.
    nR = len(r)
    nT = len(t)
    nP = len(p)

    # get the output main mesh using interpolate.
    nr_out = round(nR * fac)
    nt_out = round(nT * fac)
    np_out = round(nP * fac)

    # 1d interpolation.
    r_interp = interp1d(np.arange(nR), r, fill_value="extrapolate")
    t_interp = interp1d(np.arange(nT), t, fill_value="extrapolate")
    p_interp = interp1d(np.arange(nP), p, fill_value="extrapolate")

    r_new = r_interp(np.arange(nr_out) / (nr_out - 1) * (nR - 1))
    t_new = t_interp(np.arange(nt_out) / (nt_out - 1) * (nT - 1))
    p_new = p_interp(np.arange(np_out) / (np_out - 1) * (nP - 1))

    # get the half mesh for this new main mesh.
    # initialize the half mesh arrays.
    r_half = np.zeros(nr_out + 1)
    t_half = np.zeros(nt_out + 1)
    p_half = np.zeros(np_out + 1)

    r_half[0] = r_new[0] - (r_new[1] - r_new[0]) / 2
    r_half[1: -1] = np.array([(np.roll(r_new, -1) + r_new)[: nr_out - 1] / 2])
    r_half[-1] = r_new[nr_out - 1] + (r_new[nr_out - 1] - r_new[nr_out - 2]) / 2

    t_half[0] = t_new[0] - (t_new[1] - t_new[0]) / 2
    t_half[1: -1] = np.array([(np.roll(t_new, -1) + t_new)[: nt_out - 1] / 2])
    t_half[-1] = t_new[nt_out - 1] + (t_new[nt_out - 1] - t_new[nt_out - 2]) / 2

    p_half[1: -1] = np.array([(np.roll(p_new, -1) + p_new)[: np_out - 1] / 2])
    p_half[0] = p_half[-2] - 2 * np.pi
    p_half[-1] = p_half[1] + 2 * np.pi

    # write 1d hdf files.
    wrhdf_1d(hdf_filename=os.path.join(outdir, 'mesh_' + str(mesh_name) + '_main_r.hdf'), x=[], f=r_new)
    wrhdf_1d(hdf_filename=os.path.join(outdir, 'mesh_' + str(mesh_name) + '_main_t.hdf'), x=[], f=t_new)
    wrhdf_1d(hdf_filename=os.path.join(outdir, 'mesh_' + str(mesh_name) + '_main_p.hdf'), x=[], f=p_new)

    # write 1d half mesh hdf files.
    wrhdf_1d(hdf_filename=os.path.join(outdir, 'mesh_' + str(mesh_name) + '_half_r.hdf'), x=[], f=r_half)
    wrhdf_1d(hdf_filename=os.path.join(outdir, 'mesh_' + str(mesh_name) + '_half_t.hdf'), x=[], f=t_half)
    wrhdf_1d(hdf_filename=os.path.join(outdir, 'mesh_' + str(mesh_name) + '_half_p.hdf'), x=[], f=p_half)





""" UI Mesh python modules.
The MainWindow module: tkinter_main_app.py
The FirstWindow: upload_magnetogram.py

To start the Main UI:

from mesh_generator.ui.upload_magnetogram import MeshGeneratorUI
>>> MeshGeneratorUI()
"""

from mesh_generator.ui.upload_magnetogram import MeshGeneratorUI



"""Legacy Mesh

Legacy mesh - (step 3 of building a mesh). The legacy mesh preserves the mesh requirements including the mesh
domain, phi_shift, BG_RATIO, and legacy segment list. The legacy mesh segment list contains the mesh segment total
number of points, ratio, and domain. For more details about legacy mesh segments go to src/legacy_mesh_segment.py
"""

import json
import numpy
from mesh_generator.src.legacy_mesh_segment import LegacyMeshSegment
from mesh_generator.src.mesh_segment import MeshSegment


class LegacyMesh:
    def __init__(self, lower_bnd, upper_bnd, periodic=False, phi_shift=0.0):
        self.lower_bnd = numpy.float64(lower_bnd)
        self.upper_bnd = numpy.float64(upper_bnd)
        self.periodic = periodic
        self.phi_shift = numpy.float64(phi_shift)
        self.BG_RATIO = MeshSegment.DEFAULT_BG_REGION_RATIO
        self.segment_list = numpy.array([], dtype=LegacyMeshSegment)

    def __str__(self):
        return json.dumps(
            self.json_dict(), indent=2, default=lambda o: o.json_dict())

    def json_dict(self):
        return {
            'lower_bnd': self.lower_bnd,
            'upper_bnd': self.upper_bnd,
            'periodic': self.periodic,
            'phi_shift': self.phi_shift,
            'BG_RATIO': MeshSegment.DEFAULT_BG_REGION_RATIO,
            'segment_list': [segment.json_dict() for segment in self.segment_list.tolist()]
        }

    def insert_mesh_segment(self, new_segment, beg_index=0):
        """
        Inserts a mesh segment into the list of mesh segments and returns its
        position in the list. The mesh segments in the list are sorted by the
        start position followed by the end position in ascending order.

        Setting `beg_index` allows the comparison and insertion to start at
        `beg_index` instead of the beginning of the list, which may speed up
        the insertion process. However, it can break the sorting order of the
        mesh segments in the list if used without care.
        """
        for index in range(beg_index, len(self.segment_list)):
            segment = self.segment_list[index]
            if new_segment.s0 < segment.s0:
                break
            elif new_segment.s0 == segment.s0:
                if new_segment.s1 > segment.s1:
                    index += 1
                break
        else:
            index = len(self.segment_list)

        if index == -1:
            self.segment_list = numpy.append(self.segment_list, new_segment)
        else:
            self.segment_list = numpy.insert(
                self.segment_list, index, new_segment)

        return index



"""

Usage:
In this subpackage are the most important mesh calculations.
Returns adjusted and legacy mesh (step 2 and 3 of creating a mesh).
For more details: https://q.predsci.com/docs/mesh_generator/

Important modules:
mesh.py
mesh_segment.py

1D Mesh example:
https://q.predsci.com/docs/mesh_generator/example1d/
"""

from mesh_generator.src.mesh_segment import MeshSegment
from mesh_generator.src.mesh import Mesh


"""Mesh Optimization tool

adding or subtracting points from legacy mesh total number of points which will move the mesh up or down.
This is done iteratively.

- Checks if the mesh exceeds the user constant mesh segment requirements - if so, it will add points to the mesh to
lower its ds. It will do so by editing the file "output02_mesh_t.dat" and rerunning the fortran code to get a
new tmp_mesh_r.dat.
"""


from mesh_generator.bin.psi_mesh_tool import write_mesh_res_file


def check_seg(mesh_adj: dict, mesh_res_path: str):
    """
    This test will check if the mesh is below all constant mesh segments that the user specified.
    It will save all the mesh segments with ds=1 (constant segments) and will read mesh results to see if points are
    below segment requirements.
    Iteratively.
    :param mesh_res_path: path to mesh points file.
    :param mesh_adj: dictionary of adjusted mesh segments.
    :return: bool True/False
    """
    epsilon = 10. ** (-10.)
    exceed_seg = False

    new_constant_list = []  # a list with all constant segments
    phi_shift = mesh_adj['phi_shift']
    for segments in mesh_adj['segment_list']:
        if segments['ratio'] == 1.:
            new_constant_list.append(segments)

    idx = 0
    with open(mesh_res_path, "r") as f:
        while idx < len(new_constant_list):
            s0 = (new_constant_list[idx]['s0']) + phi_shift
            s1 = (new_constant_list[idx]['s1']) + phi_shift
            ds = (new_constant_list[idx]['ds0'])
            for i, line in enumerate(f):
                if idx == 0 and i == 0:  # very first line of the tmp_mesh_r.dat skip.
                    i += 1
                else:
                    x = float(line.split()[1:2][0])  # s
                    y = float(line.split()[2:3][0])  # ds
                    if s0 < x < s1:
                        if y > ds + epsilon:
                            # mesh is above user requests! -> break then call add_num.
                            exceed_seg = True
                            return exceed_seg
                    if x > s1:
                        # break -> move to the next constant mesh segment.
                        break
            idx += 1
    return exceed_seg


def check_ratio(adj_mesh: dict, file_name: str):
    """
    Check if the the new legacy mesh exceeds 100.5% of BG_RATIO
    :param file_name: name of file with mesh points. Ex: tmp_mesh_r.dat
    :param adj_mesh: dictionary with adjusted mesh results.
    :return: bool True/False
    """
    percent = 0.005
    exceed_ratio = False
    ratio_max = adj_mesh['BG_RATIO'] * (1 + percent)
    ratio_min = (1. / adj_mesh['BG_RATIO']) * (1 - percent)
    with open(str(file_name), "r") as f:
        for idx, line in enumerate(f):
            if idx >= 1:
                for word in line.split()[3:4]:
                    if float(word) > ratio_max:
                        exceed_ratio = True
                        # print("CURRENT RATIO " + str(word) + " > " + str(ratio_max) + " BG_RATIO")
                        return exceed_ratio
                    if float(word) < ratio_min:
                        exceed_ratio = True
                        # print("CURRENT RATIO " + str(word) + " < " + str(ratio_min) + " 1/BG_RATIO")
                        return exceed_ratio
    return exceed_ratio


def adjust_txt_file(name_of_file: str, replace_num: int):
    """
    :param replace_num: adjusted number of points in legacy mesh
    :param name_of_file: for example, output02_mesh_t.dat
    :return: a modified output02_mesh_t.dat with num_of_points -1 / +1
    """

    with open(name_of_file, 'r') as f:
        file_data = f.readlines()

    # Replace the total number of points line #5 in output02_mesh_p.dat
    file_data[5] = str(replace_num) + '\n'

    # # Write the file out again
    with open(name_of_file, 'w') as f:
        f.writelines(file_data)


def reduce_num(adj_mesh: dict, total_num: int, mesh_res_path: str, output_path: str):
    """
    Iteratively find the optimized number of points for mesh by reducing the total number of points (-1).
    :param output_path: file with legacy mesh info.
    :param mesh_res_path: file with mesh points.
    :param total_num: total number of points in legacy mesh.
    :param adj_mesh: dictionary with adjusted mesh segments.
    :return: now the tmp_mesh_r.dat is optimal.
    """
    # print("Initial total number of points:", total_num)
    idx = 0

    if check_ratio(adj_mesh, mesh_res_path) is False:
        while idx <= int(total_num):
            total_num += - 1
            adjust_txt_file(output_path, total_num)  # modify output02_mesh_t.dat
            write_mesh_res_file(mesh_res_path, output_path)

            if check_seg(adj_mesh, mesh_res_path) is False and \
                    check_ratio(adj_mesh, mesh_res_path) is False:
                idx += 1
            else:
                # print("Final num_of_points is", total_num + 1)
                adjust_txt_file(output_path, (total_num + 1))
                break
    # else:
    # print("Final num_of_points is ", total_num)


def add_num(adj_mesh: dict, total_num: int, mesh_res_path: str, output_path: str):
    """
    Mesh is exceeding the user requests more points are needed (+1).
    :param output_path: name of output file with mesh characteristics.
    :param mesh_res_path: name of file with mesh points.
    :param total_num: total number of points in legacy mesh.
    :param adj_mesh: dictionary of adjusted mesh segments.
    """
    # print("Initial total number of points:", total_num)

    total_num = total_num + 1  # increase total number of points.

    adjust_txt_file(output_path, total_num)  # modify output02_mesh_t.dat (+1 total number of points).
    write_mesh_res_file(mesh_res_path, output_path)
    idx = 0

    while idx <= int(total_num):
        if check_seg(adj_mesh, mesh_res_path) is True:
            total_num += 1
            adjust_txt_file(output_path, total_num)  # modify output02_mesh_t.dat (+1 total number of points).
            write_mesh_res_file(mesh_res_path, output_path)
            idx += 1
        else:
            # print("Final num_of_points is", total_num)
            break


def apply_filter(output_path: str, num_filter: int):
    """ Function to adjust the output.dat last line to be 5. """
    with open(output_path, "r") as file:
        # Read a list of lines into data
        data = file.readlines()

    # Now change the last line.
    data[-1] = str(num_filter)

    # Write everything back
    with open(output_path, 'w') as file2:
        file2.writelines(data)


def check_mesh_valid(adj_mesh: dict, total_legacy_num: int, output_path, mesh_res_path):
    if check_seg(adj_mesh, mesh_res_path):
        # print("Original mesh is above user requests, calling add_num.")  # needs to add points to lower the mesh (+1).
        add_num(adj_mesh, total_legacy_num, mesh_res_path, output_path)
    else:  # need to optimize your mesh by deleting points (-1).
        # print("Original mesh is below user requests, calling reduce_num to optimize mesh results.")
        reduce_num(adj_mesh, total_legacy_num, mesh_res_path, output_path)

    # print("Mesh is valid. Applying filter = 5 to smooth mesh.")
    apply_filter(output_path, 5)
    write_mesh_res_file(mesh_res_path, output_path)




""" Read in mesh_res.dat files with mesh 1D points, resolution, and cell to cell ratio. """
import numpy as np
import subprocess


def read_tmp_file(name_of_tmp_file: str):
    """Read mesh_res_r/t/p.dat in mesh_generator/bin with mesh points. This will be used to write hdf visc file.
    :param name_of_tmp_file: str. name of the tmp file.
    :return: s, ds, ratio
    """
    total_num_of_points = file_len(name_of_tmp_file) - 1
    s = np.zeros(total_num_of_points)
    ds = np.zeros(total_num_of_points)
    ratio = np.zeros(total_num_of_points)

    with open(str(name_of_tmp_file), 'r') as f:
        for idx, line in enumerate(f):
            if idx >= 1:
                vec_line = line.split()
                s[idx - 1] = float(vec_line[1])
                ds[idx - 1] = float(vec_line[2])
                ratio[idx - 1] = float(vec_line[3])
    return s, ds, ratio


def file_len(file_name):
    """ Returns the number of lines in a file. """
    p = subprocess.Popen(['wc', '-l', file_name], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    result, err = p.communicate()
    if p.returncode != 0:
        raise IOError(err)
    return int(result.strip().split()[0])



"""
plotting the mesh segments as inputted and after resolve_mesh_segment function is applied
which includes these 3 functions:
    1.    _resolve_mesh_segments_overlap_region()
    2.    _resolve_mesh_segments_narrow_region()
    3.    _resolve_mesh_segments_ds()
"""
from matplotlib import pyplot as plt


def plot_segments(inputed_mesh, adjusted_mesh):
    fig, axs = plt.subplots(3)
    '''
    plot the input segments- given by the user
    '''

    for segments in inputed_mesh['segment_list']:
        axs[0].hlines(segments['ds0'], segments['s0'], segments['s1'], linewidth=3)
        axs[0].axvline(x=segments['s1'], dashes=[6, 2], color='gray')
        axs[0].axvline(x=segments['s0'], dashes=[6, 2], color='gray')
    axs[0].set_title('Input segments')
    axs[0].set(xlabel=' \u03B8 ', ylabel='\u0394 \u03B8')

    '''
    plot the output segments- after applying the function called "resolve_mesh_segments()"
    '''

    for segments in adjusted_mesh['segment_list']:
        axs[1].hlines(max(segments['ds0'], segments['ds1']), segments['s0'], segments['s1'], linewidth=3)
        axs[1].axvline(x=segments['s1'], dashes=[6, 2], color='gray')
        axs[1].axvline(x=segments['s0'], dashes=[6, 2], color='gray')
    axs[1].set_title('_resolve_mesh_segments_overlap_region()')
    axs[1].set(xlabel=' \u03B8 ', ylabel='\u0394 \u03B8')

    '''
    plot the output segments- after applying the function called "resolve_mesh_segments()"- 
    linear line connecting 2 points
    
    '''

    for segments in adjusted_mesh['segment_list']:
        axs[2].plot([segments['s1'], segments['s0']], [segments['ds1'], segments['ds0']], linewidth=3, color='k')
        axs[2].axvline(x=segments['s1'], dashes=[6, 2], color='gray')
        axs[2].axvline(x=segments['s0'], dashes=[6, 2], color='gray')
        axs[2].set_title('_resolve_mesh_segments_ds()')
        axs[2].set(xlabel=' \u03B8 ', ylabel='\u0394 \u03B8')

    plt.tight_layout()
    plt.show()


if __name__ == "__main__":
    from tests.ar_test import *

    my_mesh = get_mesh_phi_1().json_dict()
    adjusted_mesh = adjust__mesh_phi_1().json_dict()
    plot_segments(my_mesh, adjusted_mesh)



""" Add a vertical scroll bar to frame. This is used for tk/tkinter_mainapp.py"""
import tkinter as tk


class VerticalScrolledFrame:
    """
    A vertically scrolled Frame that can be treated like any other Frame
    ie it needs a master and layout and it can be a master.
    :width:, :height:, :bg: are passed to the underlying Canvas
    :bg: and all other keyword arguments are passed to the inner Frame
    note that a widget layed out in this frame will have a self.master 3 layers deep,
    (outer Frame, Canvas, inner Frame) so
    if you subclass this there is no built in way for the children to access it.
    You need to provide the controller separately.
    """

    def __init__(self, master, **kwargs):
        width = kwargs.pop('width', None)
        height = kwargs.pop('height', None)
        bg = kwargs.pop('bg', kwargs.pop('background', None))
        self.outer = tk.Frame(master, **kwargs)

        self.vsb = tk.Scrollbar(self.outer, orient=tk.VERTICAL)
        self.vsb.pack(fill=tk.Y, side=tk.RIGHT)
        self.canvas = tk.Canvas(self.outer, highlightthickness=0, width=width, height=height, bg=bg)
        self.canvas.pack(side=tk.LEFT, fill=tk.BOTH, expand=True)
        self.canvas['yscrollcommand'] = self.vsb.set
        # mouse scroll does not seem to work with just "bind"; You have
        # to use "bind_all". Therefore to use multiple windows you have
        # to bind_all in the current widget
        self.canvas.bind("<Enter>", self._bind_mouse)
        self.canvas.bind("<Leave>", self._unbind_mouse)
        self.vsb['command'] = self.canvas.yview

        self.inner = tk.Frame(self.canvas, bg=bg)
        # pack the inner Frame into the Canvas with the topleft corner 4 pixels offset
        self.canvas.create_window(4, 4, window=self.inner, anchor='nw')
        self.inner.bind("<Configure>", self._on_frame_configure)

        self.outer_attr = set(dir(tk.Widget))

    def __getattr__(self, item):
        if item in self.outer_attr:
            # geometry attributes etc (eg pack, destroy, tkraise) are passed on to self.outer
            return getattr(self.outer, item)
        else:
            # all other attributes (_w, children, etc) are passed to self.inner
            return getattr(self.inner, item)

    def _on_frame_configure(self, event=None):
        x1, y1, x2, y2 = self.canvas.bbox("all")
        height = self.canvas.winfo_height()
        self.canvas.config(scrollregion=(0, 0, x2, max(y2, height)))

    def _bind_mouse(self, event=None):
        self.canvas.bind_all("<4>", self._on_mousewheel)
        self.canvas.bind_all("<5>", self._on_mousewheel)
        self.canvas.bind_all("<MouseWheel>", self._on_mousewheel)

    def _unbind_mouse(self, event=None):
        self.canvas.unbind_all("<4>")
        self.canvas.unbind_all("<5>")
        self.canvas.unbind_all("<MouseWheel>")

    def _on_mousewheel(self, event):
        """Linux uses event.num; Windows / Mac uses event.delta"""
        if event.num == 4 or event.delta > 0:
            self.canvas.yview_scroll(-1, "units")
        elif event.num == 5 or event.delta < 0:
            self.canvas.yview_scroll(1, "units")


#  **** SCROLL BAR TEST *****
if __name__ == "__main__":
    root = tk.Tk()
    root.title("Scrollbar Test")
    root.geometry('400x500')

    frame = VerticalScrolledFrame(root,
                                  width=300,
                                  borderwidth=2,
                                  relief=tk.SUNKEN,
                                  background="light gray")
    # frame.grid(column=0, row=0, sticky='nsew') # fixed size
    frame.pack(fill=tk.BOTH, expand=True)  # fill window

    for i in range(30):
        label = tk.Label(frame, text="This is a label " + str(i))
        label.grid(column=1, row=i, sticky=tk.W)

        text = tk.Entry(frame, textvariable="text")
        text.grid(column=2, row=i, sticky=tk.W)

    root.mainloop()



"""
Write output02_mesh_t/p/r.dat file with legacy mesh results.
"""


def write_output_file(legacy_mesh: dict, NUM_OF_FILTER: int, mesh_type: str, output_path: str):
    """
    Print the solution of the legacy test in a output02_mesh_t.dat file where it will be
    similar to the IDL file format so the fortran mesh will be able to read the mesh requirements and
    generate a mesh_res.txt file with all the mesh points


    Arguments:
        :param output_path: full path to the output file written.
        :param mesh_type: 't'/'p'/'r'
        :param legacy_mesh : dictionary of legacy mesh
        :param NUM_OF_FILTER: number of filters applied
    """

    tol_n = 0  # total number of points in legacy mesh
    ratio_in_s = []  # ratio between s0 and legacy upper bound
    ratio_in_ds = []  # ration between ds0 and ds1

    for idx in range(0, len(legacy_mesh['segment_list'])):
        segments = legacy_mesh['segment_list'][idx]
        tol_n += segments["num_points"]
        ratio_in_s.append((segments["s0"] - legacy_mesh["lower_bnd"]) /
                          (legacy_mesh["upper_bnd"] - legacy_mesh["lower_bnd"]))
        if idx == len(legacy_mesh['segment_list']) - 1:
            ratio_in_s.append(segments["s1"] / legacy_mesh["upper_bnd"])
        ratio_in_ds.append(segments["ratio"])

    # open .dat file
    with open(output_path, "w") as f:
        f.write("Mesh coordinate label:\n%s \n" % mesh_type)
        f.write("Flag to specify a periodic mesh [.true.|.false.]:\n%s \n" % _is_periodic(legacy_mesh['periodic']))
        f.write("Number of mesh points:\n%s \n" % tol_n)
        f.write(
            "Lower and upper mesh limits [X0,X1]:\n %s, %s \n" % (legacy_mesh["lower_bnd"], legacy_mesh["upper_bnd"]))

        if not legacy_mesh['periodic']:
            f.write("Position of segments, FRAC [terminated with /]:\n %s" % ratio_in_s[0])
            for i in range(1, len(ratio_in_s)):
                f.write(', ' + str(ratio_in_s[i]))
            f.write('/ \n')

        if legacy_mesh['periodic']:
            f.write("Position of segments, FRAC [terminated with /]:\n %s" % ratio_in_s[0])
            for i in range(1, len(ratio_in_s) - 1):
                f.write(', ' + str(ratio_in_s[i]))
            f.write('/ \n')

        f.write("Ratio of segments, DRATIO [terminated with /]:\n %s" % (ratio_in_ds[0]))
        for i in range(1, len(ratio_in_ds)):
            f.write(', ' + str(ratio_in_ds[i]))
        f.write('/ \n')

        f.write("Amount to shift the mesh:\n %s \n" % legacy_mesh['phi_shift'])

        f.write("Number of times to filter:\n%s \n" % NUM_OF_FILTER)

    return tol_n


def _is_periodic(legacy):
    if not legacy:
        return ".false."
    else:
        return ".true."


if __name__ == "__main__":
    from tests.ar_test import *

    write_output_file(legacy__mesh_phi_2().json_dict())



"""
A function "get_num_points()" to return the total number of points in t/p/r mesh. The total number of points
is found in bin/output.dat file.
"""


def get_num_points(file_path):
    """
    Return the total number of mesh points. Line 5 in bin.output.dat file.
    """
    with open(file_path, 'r') as res:
        num_points = res.readlines()[5]

    return num_points



"""
User Interface (UI) main window to create a mesh generation by uploading a hdf file magnetogram
plotting it using matplotlib and interactively specifying the mesh requirement. This is a great way to develop more
tests for test/unit_test.py and detect bugs in the src mesh code.
"""

import tkinter as tk
import numpy as np
import matplotlib as mpl
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg, NavigationToolbar2Tk
from matplotlib.figure import Figure
from matplotlib.widgets import RectangleSelector
from matplotlib.patches import Rectangle
from mesh_generator.bin.call_psi_mesh_tool import create_psi_mesh
from mesh_generator.src.mesh_segment import MeshSegment
from mesh_generator.src.mesh import Mesh
from mesh_generator.ui.mesh_info import MeshInfo
from mesh_generator.ui.mesh_box import MeshBox
from mesh_generator.ui.read_hdf import plot_hdf
from mesh_generator.bin.tmp_mas_template import write_tmp_mas_file
from mesh_generator.bin.mesh_header_template import write_mesh_header
from mesh_generator.hdf.read_mesh_res import read_tmp_file
from mesh_generator.hdf.write_1d_hdf_files import write_1d_mesh_hdf
from mesh_generator.ui.save_dict import SaveDict
from mesh_generator.ui.restrictor_pan import RestrictPan
from mesh_generator.ui.my_axes import My_Axes
from mesh_generator.ui.mesh_num_points import get_num_points
from mesh_generator.ui.scroll_bar import VerticalScrolledFrame
from mesh_generator.ui.table_chart import TableChart
import os


class MeshApp:
    """The User Interface main app to develop a 3D mesh. """

    def __init__(self, hdf_file_path, ui_session_data, save_mesh_file_path, UseExistingInput=False):
        root = tk.Tk()
        root.wm_title("MainApp")  # title of the window.
        root.config(background="white")  # background color of the window.
        root.geometry("1825x640")  # window size.
        #root.minsize(width=1825, height=640)  # minimum window size that the user can adjust.

        # assign the mesh results folder
        self.mesh_results_folder = save_mesh_file_path

        # frame with instructions on how to use gui includes step 1 and step 2.
        # (submitting the boxes and input respective ds.)
        self.leftFrame = VerticalScrolledFrame(root, width=440, height=600, relief=tk.SUNKEN, background="gray92")
        self.leftFrame.grid(row=0, column=0, sticky=tk.W)  # position of the frame on the main window

        # label with step 1 instruction
        tk.Label(self.leftFrame, text="Step 1: Draw a box, when box is finalized click submit box button.",
                 font=7).grid(row=0, column=0, columnspan=5, sticky=tk.NW, pady=10)

        # save latest box button
        tk.Button(self.leftFrame, text='Save latest box', activeforeground='navy',
                  command=self.save_latest_box).grid(row=1, column=1, pady=10)

        # reset latest box button
        tk.Label(self.leftFrame, text="Remove all boxes:").grid(row=2, column=0, sticky=tk.E)
        tk.Button(self.leftFrame, text='Reset', activeforeground='navy', command=self.reset) \
            .grid(row=2, column=1, pady=10)

        # delete drop-box
        tk.Label(self.leftFrame, text="Remove box #:").grid(row=3, column=0, sticky=tk.E)
        self.del_tb_variable = tk.StringVar(self.leftFrame)
        self.del_tb_variable.set(' ')
        self.del_tb_opt = tk.OptionMenu(self.leftFrame, self.del_tb_variable, ' ')
        self.del_tb_opt.grid(row=3, column=1, sticky=tk.NSEW)
        self.del_tb_variable.trace("w", self.drop_box_delete_box_get_val)
        tk.Button(self.leftFrame, text='apply', command=self.delete_box).grid(row=3, column=2, sticky=tk.W)

        # redraw drop-box
        tk.Label(self.leftFrame, text="Redraw box #:").grid(row=4, column=0, sticky=tk.E)
        self.redraw_variable = tk.StringVar(self.leftFrame)
        self.redraw_variable.set(' ')
        self.redraw_opt = tk.OptionMenu(self.leftFrame, self.redraw_variable, ' ')
        self.redraw_opt.grid(row=4, column=1, sticky=tk.NSEW)
        self.del_tb_variable.trace("w", self.redraw_drop_box_get_val)

        tk.Button(self.leftFrame, text='apply', command=self.redraw_box_apply).grid(row=4, column=2, sticky=tk.W)
        tk.Button(self.leftFrame, text='submit', command=self.redraw_submit).grid(row=4, column=3, sticky=tk.W)

        # optional edit box table manually in a new window.
        tk.Label(self.leftFrame, text="(Optional) Enter box coordinates manually").grid(row=5, column=0,
                                                                                        sticky=tk.W, columnspan=3,
                                                                                        pady=10)
        tk.Button(self.leftFrame, text='Box Chart', command=self.box_chart).grid(row=6, column=1)

        # step 2- enter ds values
        tk.Label(self.leftFrame, text="Step 2: After submitting each box indicate desired ds in textbox below.",
                 font=6).grid(row=7, column=0, columnspan=7, sticky=tk.W, pady=10)
        tk.Label(self.leftFrame, text="ds", relief=tk.RIDGE, bg='lightgray', bd=2). \
            grid(row=8, column=1, sticky=tk.NSEW)
        tk.Label(self.leftFrame, text="var_ds_ratio", relief=tk.RIDGE, bg='lightgray', bd=2). \
            grid(row=8, column=2, sticky=tk.NSEW, columnspan=2)

        # second frame: includes step 3, 4 and building a radial mesh manually.
        self.rightFrame = VerticalScrolledFrame(root, width=465, height=600, borderwidth=2, relief=tk.SUNKEN,
                                                background="gray92")
        self.rightFrame.grid(row=0, column=2)

        # step 3- enter BG ratio
        tk.Label(self.rightFrame, text="Step 3: Input BG ratio:", font=7).grid(row=0, column=0, sticky=tk.W, pady=10)
        self.bg_ratio = tk.Entry(self.rightFrame, width=5)
        self.bg_ratio.insert(tk.END, '1.06')
        self.bg_ratio.grid(row=0, column=1, sticky=tk.W)

        # enter FG ratio
        tk.Label(self.rightFrame, text="Input FG ratio:", font=7).grid(row=0, column=2, sticky=tk.E, pady=10,
                                                                       columnspan=2)
        self.fg_ratio = tk.Entry(self.rightFrame, width=5)
        self.fg_ratio.insert(tk.END, '1.03')
        self.fg_ratio.grid(row=0, column=4, sticky=tk.W)

        # step 4- get total number of points in theta and phi
        tk.Button(self.rightFrame, text='\u03B8 number of points', font=7, command=self.num_points_theta) \
            .grid(row=2, column=0, columnspan=1, pady=10)
        tk.Button(self.rightFrame, text='\u03C6 number of points', font=7, command=self.num_points_phi) \
            .grid(row=3, column=0, columnspan=1, pady=10)

        tk.Label(self.rightFrame, text="     ", font=7).grid(row=2, column=1, pady=10, sticky=tk.W)
        tk.Label(self.rightFrame, text="     ", font=7).grid(row=3, column=1, pady=10, sticky=tk.W)

        # step 4- get plots of phi and theta mesh results
        tk.Label(self.rightFrame, text="Step 4: Get mesh results", font=7).grid(row=1, column=0, sticky=tk.W,
                                                                                pady=10)
        tk.Button(self.rightFrame, text='\u03B8 mesh results', font=7, command=self.get_theta_results). \
            grid(row=2, column=2, sticky=tk.W, columnspan=2, pady=10)
        tk.Button(self.rightFrame, text='\u03C6 mesh results', font=7, command=self.get_phi_results). \
            grid(row=3, column=2, sticky=tk.W, columnspan=2, pady=10)

        # step 5- radial mesh table
        tk.Label(self.rightFrame, text="Step 5: Input radial mesh segments", font=7).grid(row=4, column=0,
                                                                                          columnspan=2, pady=10,
                                                                                          sticky=tk.W)
        # get radial total number of mesh points
        tk.Button(self.rightFrame, text='r number of points', font=7, command=self.num_points_r). \
            grid(row=5, column=0, columnspan=1, pady=10)
        tk.Label(self.rightFrame, text="     ", font=7).grid(row=5, column=1, pady=10, sticky=tk.W)

        # get radial mesh resulted plot
        tk.Button(self.rightFrame, text='r mesh results', font=7, command=self.get_radial_results). \
            grid(row=5, column=2, sticky=tk.W, columnspan=2, pady=10)

        # total number of points of all mesh (t x p x r):
        tk.Label(self.rightFrame, text="Compute total number of points:", font=7).grid(row=6, column=0, pady=10,
                                                                                       sticky=tk.W, columnspan=2)
        tk.Button(self.rightFrame, text='\u03B8 x \u03C6 x r', font=7, command=self.total_num_points). \
            grid(row=6, column=2, sticky=tk.NSEW, pady=10)

        # RADIAL_BG_TEXTBOX
        tk.Label(self.rightFrame, text="Radial BG ratio:", font=7).grid(row=7, column=0, sticky=tk.E)
        self.radial_bg_ratio = tk.Entry(self.rightFrame, width=5)
        self.radial_bg_ratio.insert(tk.END, '1.06')
        self.radial_bg_ratio.grid(row=7, column=1, sticky=tk.W)

        # RADIAL_FG_TEXTBOX
        tk.Label(self.rightFrame, text="Radial FG ratio:", font=7).grid(row=7, column=2, sticky=tk.E, columnspan=2)
        self.radial_fg_ratio = tk.Entry(self.rightFrame, width=5)
        self.radial_fg_ratio.insert(tk.END, '1.03')
        self.radial_fg_ratio.grid(row=7, column=4, sticky=tk.W)

        # R0_TEXTBOX
        tk.Label(self.rightFrame, text="rMin:", font=7).grid(row=8, column=0, sticky=tk.E)
        self.rMin = tk.Entry(self.rightFrame, width=5)
        self.rMin.insert(tk.END, '1')
        self.rMin.grid(row=8, column=1, sticky=tk.W)

        # R1_TEXTBOX
        tk.Label(self.rightFrame, text="rMax:", font=7).grid(row=9, column=0, sticky=tk.E)
        self.rMax = tk.Entry(self.rightFrame, width=5)
        self.rMax.insert(tk.END, '2.5')
        self.rMax.grid(row=9, column=1, sticky=tk.W)

        # # RADIAL TABLE
        tk.Label(self.rightFrame, text='r0', relief=tk.RIDGE, bg='lightgray', bd=2, width=6).grid(row=10, column=1,
                                                                                                  sticky=tk.NSEW,
                                                                                                  columnspan=1)
        tk.Label(self.rightFrame, text='r1', relief=tk.RIDGE, bg='lightgray', bd=2, width=6).grid(row=10, column=2,
                                                                                                  sticky=tk.NSEW,
                                                                                                  columnspan=1)
        tk.Label(self.rightFrame, text='ds', relief=tk.RIDGE, bg='lightgray', bd=2, width=6).grid(row=10, column=3,
                                                                                                  sticky=tk.NSEW,
                                                                                                  columnspan=1)
        tk.Label(self.rightFrame, text='var_ds_ratio', relief=tk.RIDGE, bg='lightgray', bd=2, width=9).grid(row=10,
                                                                                                            column=4,
                                                                                                            sticky=tk.NSEW,
                                                                                                            columnspan=1)

        # save patch of rect drawn on canvas
        self.rect_data = 0
        # save all boxes drawn on the plot in a list
        self.patch_list = []
        # if the button redraw rectangle is pressed then keep track if the new rectangle is drawn on plot.
        self.redraw_box_submit = False
        # if the box number passed in redraw textbox
        self.box_num = 0
        # ds textbox widgets saved in a list
        self.ds_tab_array = []
        # var ds textbox widgets saved in a list
        self.var_ds_tab_array = []
        # radial textbox specifications saved in a list
        self.radial_table_tb = []

        # pan only left and right
        mpl.projections.register_projection(My_Axes)

        # frame containing the canvas
        self.midFrame = tk.Frame(root, width=300, height=200, background="white")
        self.midFrame.grid(row=0, column=1, rowspan=1)
        # Make the canvas decently large.
        self.fig = Figure(figsize=(9, 4.5), dpi=100)
        self.fig.subplots_adjust(left=0.02, bottom=0.1, right=0.96, top=0.92, wspace=0, hspace=0)
        self.ax = self.fig.add_subplot(111, projection="My_Axes")
        self.hdf_file_path = hdf_file_path
        cmin = self.setup_magnetogram_plot(self.fig, self.hdf_file_path)

        # save ui_session button.
        self.bottomFrame = tk.Frame(root, width=1830, height=10)
        self.bottomFrame.grid(row=1, column=0, columnspan=3)
        tk.Label(self.bottomFrame, text="\t".expandtabs(195), font=7).grid(row=0, column=0, sticky=tk.NSEW)
        tk.Button(self.bottomFrame, text="Save session", font=7, command=self.save_ui_session).grid(row=0, column=1)
        tk.Label(self.bottomFrame, text="\t".expandtabs(25), font=7).grid(row=0, column=5, sticky=tk.NSEW)
        tk.Label(self.bottomFrame, text="vmin/vmax:", font=7).grid(row=0, column=9)
        self.cmax_value = tk.Entry(self.bottomFrame, width=5)
        self.cmax_value.insert(tk.END, str(abs(cmin)))
        self.cmax_value.grid(row=0, column=10, sticky=tk.W)
        tk.Button(self.bottomFrame, text="apply", font=7, command=self.apply_new_scale).grid(row=0, column=11)
        tk.Label(self.bottomFrame, text="\t".expandtabs(195), font=7).grid(row=0, column=12, sticky=tk.NSEW)

        self.canvas = FigureCanvasTkAgg(self.fig, master=self.midFrame)
        self.canvas.draw()
        self.canvas.get_tk_widget().pack(side=tk.BOTTOM, fill=tk.BOTH, expand=1, anchor=tk.CENTER)

        def line_select_callback(eclick, erelease):
            """ draw rectangle boxes on matplotlib figure """
            x1, y1 = eclick.xdata, eclick.ydata
            x2, y2 = erelease.xdata, erelease.ydata
            rect = Rectangle((min(x1, x2), min(y1, y2)), np.abs(x1 - x2), np.abs(y1 - y2), fill=False)
            self.rect_data = rect

        rs = RectangleSelector(self.ax, line_select_callback,
                               drawtype='box', useblit=False, button=[1],
                               minspanx=5, minspany=5, spancoords='pixels',
                               interactive=True)

        toolbar = NavigationToolbar2Tk(self.canvas, self.midFrame)
        toolbar.update()
        # canvas._tkcanvas.pack(side=tk.TOP)

        # read dictionary / MeshInfo object
        if UseExistingInput is False:
            print("new ui_session!")
            self.ui_session = ui_session_data
        if UseExistingInput is True:
            print("uploading json!")
            self.read_ui_session(ui_session_data)

        root.mainloop()

    def read_ui_session(self, ui_session_data):
        # copy information from dictionary to MeshInfo object that can later be edited & saved while using the UI.
        self.ui_session = MeshInfo(lower_bnd_t=ui_session_data["lower_bnd_t"],
                                   upper_bnd_t=ui_session_data["upper_bnd_t"],
                                   lower_bnd_p=ui_session_data["lower_bnd_p"],
                                   upper_bnd_p=ui_session_data["upper_bnd_p"],
                                   lower_bnd_r=ui_session_data["lower_bnd_r"],
                                   upper_bnd_r=ui_session_data["upper_bnd_r"],
                                   BG_RATIO=ui_session_data["BG_RATIO"],
                                   FG_RATIO=ui_session_data["FG_RATIO"],
                                   RADIAL_BG_RATIO=ui_session_data["RADIAL_BG_RATIO"],
                                   RADIAL_FG_RATIO=ui_session_data["RADIAL_FG_RATIO"])

        if len(ui_session_data["box_list"]) != 0:
            self.ui_session.box_list = ui_session_data["box_list"]

        i = 0
        while i < len(self.ui_session.box_list):
            box = self.ui_session.box_list[i]
            self.ui_session.box_list[i] = MeshBox(t0=box['t0'], t1=box['t1'], p0=box['p0'],
                                                  p1=box['p1'], r0=box['r0'], r1=box['r1'], ds=box['ds'],
                                                  radial_ds=box['radial_ds'], radial_var_ds=box['radial_var_ds'],
                                                  tp_var_ds=box['tp_var_ds'])
            i += 1

        # draw boxes on the image
        i = 1
        for rect in self.ui_session.box_list:
            rect_patch = mpl.patches.Rectangle((min(rect.p0, rect.p1), min(rect.t0, rect.t1)),
                                               np.abs(rect.p0 - rect.p1),
                                               np.abs(rect.t0 - rect.t1), fill=False)
            label_box = "Box # %s" % i
            self.ax.text(rect.p0, rect.t0, label_box)
            self.ax.add_patch(rect_patch)
            self.patch_list.append(rect_patch)

            # ds text box
            tk.Label(self.leftFrame, text=label_box, relief=tk.RIDGE, bg='lightgray', bd=2). \
                grid(row=8 + i, column=0, sticky=tk.NSEW)
            self.ds_tab_array.append(tk.Entry(self.leftFrame, width=11, bg='white'))
            self.var_ds_tab_array.append(tk.Entry(self.leftFrame, width=11, bg='white'))
            self.ds_tab_array[-1].insert(tk.END, str(rect.ds))
            if rect.tp_var_ds is not None:
                self.var_ds_tab_array[-1].insert(tk.END, str(rect.tp_var_ds))
            else:
                self.var_ds_tab_array[-1].insert(tk.END, '')
            self.ds_tab_array[-1].grid(row=8 + i, column=1, sticky=tk.NSEW)
            self.var_ds_tab_array[-1].grid(row=8 + i, column=2, sticky=tk.NSEW, columnspan=2)

            # add r table with textbox
            tk.Label(self.rightFrame, text=label_box, relief=tk.RIDGE, bg='lightgray', bd=2) \
                .grid(row=10 + i, column=0, sticky=tk.NSEW)
            cols = []
            for j in range(4):
                e = tk.Entry(self.rightFrame, width=5)
                if j == 0 and rect.r0 is not None:
                    e.insert(tk.END, str(rect.r0))
                elif j == 1 and rect.r1 is not None:
                    e.insert(tk.END, str(rect.r1))
                elif j == 2 and rect.radial_ds is not None:
                    e.insert(tk.END, str(rect.radial_ds))
                elif j == 3 and rect.radial_var_ds is not None:
                    e.insert(tk.END, str(rect.radial_var_ds))
                else:
                    e.insert(tk.END, '')
                e.grid(row=10 + i, column=j + 1, sticky=tk.NSEW)
                cols.append(e)
            self.radial_table_tb.append(cols)
            i += 1

        # add BG_RATIO value to textbox
        self.bg_ratio.delete(0, tk.END)
        self.bg_ratio.insert(0, str(self.ui_session.BG_RATIO))

        # add FG_RATIO value to textbox
        self.fg_ratio.delete(0, tk.END)
        self.fg_ratio.insert(0, str(self.ui_session.FG_RATIO))

        # add BG_RATIO_RADIAL value to textbox
        self.radial_bg_ratio.delete(0, tk.END)
        self.radial_bg_ratio.insert(0, str(self.ui_session.RADIAL_BG_RATIO))

        # add FG_RATIO_RADIAL value to textbox
        self.radial_fg_ratio.delete(0, tk.END)
        self.radial_fg_ratio.insert(0, str(self.ui_session.RADIAL_FG_RATIO))

        # add rMin value to textbox
        self.rMin.delete(0, tk.END)
        self.rMin.insert(0, str(self.ui_session.lower_bnd_r))

        # add rMax value to textbox
        self.rMax.delete(0, tk.END)
        self.rMax.insert(0, str(self.ui_session.upper_bnd_r))

        # add drop-box for deleting mesh boxes, options.
        self.remove_menu_option_delete_dropbox()
        self.remove_menu_option_redraw_dropbox()

    def reset(self):
        """ delete all the boxes drawn on matplotlib figure and clear ui_session box_list."""
        for i in range(0, len(self.ui_session.box_list)):
            # delete the lastest ds textbox and label
            tk.Label(self.leftFrame, text="             ").grid(row=9 + i,
                                                                column=0, sticky=tk.NSEW)
            self.ds_tab_array[i].destroy()
            self.var_ds_tab_array[i].destroy()

            # delete box label in radial table
            tk.Label(self.rightFrame, text="       ", bg='gray92', bd=2).grid(row=11 + i, column=0, sticky=tk.NSEW)

        # delete the entry boxes in radial table
        for row in self.radial_table_tb:
            for entry in row:
                entry.destroy()

        self.radial_table_tb = []
        self.ds_tab_array = []
        self.var_ds_tab_array = []
        self.ds_tab_array = []
        # delete all boxes in dictionary
        self.ui_session.box_list = np.array([], dtype=MeshBox)

        # remove rectangles from plot
        for box in self.patch_list:
            box.remove()
        self.patch_list.clear()

        # delete all text on plot
        self.ax.texts.clear()

        # remove all dropbox options
        self.reset_delete_box_drop_box()
        self.reset_redraw_box_drop_box()

    def num_points_theta(self):
        """ return the total number of points in current theta mesh. """
        self.get_bg_ratio()
        self.get_fg_ratio()
        self.add_ds()
        try:
            self.get_mesh_theta_1(False)
            num_points_t = get_num_points(os.path.join(self.mesh_results_folder, "tmp_mesh_t.dat"))
        except:
            num_points_t = "n/a"
        tk.Label(self.rightFrame, text=str(num_points_t), font=7).grid(row=2, column=1, sticky=tk.SW)
        return num_points_t

    def num_points_phi(self):
        """ return the total number of points in current phi mesh. """
        self.get_bg_ratio()
        self.get_fg_ratio()
        self.add_ds()
        try:
            self.get_mesh_phi_1(False)
            num_points_p = get_num_points(os.path.join(self.mesh_results_folder, "tmp_mesh_p.dat"))
        except:
            num_points_p = "n/a"
        tk.Label(self.rightFrame, text=str(num_points_p), font=7).grid(row=3, column=1, sticky=tk.SW)
        return num_points_p

    def num_points_r(self):
        """ return the total number of points in current radial mesh. """
        self.get_radial_bg_ratio()
        self.get_radial_rMin()
        self.get_radial_rMax()
        try:
            self.get_mesh_r_1(False)
            num_points_r = get_num_points(os.path.join(self.mesh_results_folder, "tmp_mesh_r.dat"))
        except:
            num_points_r = "n/a"
        tk.Label(self.rightFrame, text=str(num_points_r), font=7).grid(row=5, column=1, sticky=tk.SW)
        return num_points_r

    def total_num_points(self):
        """ return the total number of points in 3d mesh: np_t x np_p x np_r """
        try:
            np_t = self.num_points_theta()
            np_p = self.num_points_phi()
            np_r = self.num_points_r()
            total_np = int(np_t) * int(np_p) * int(np_r)
        except:
            total_np = "n/a"
        tk.Label(self.rightFrame, text=str('{:.2e}'.format(total_np)), font=6).grid(row=6, column=3, sticky=tk.W,
                                                                                    columnspan=2)

    def write_tmp_mas_files(self):
        """write a tmp mas files"""
        all_tmp_mesh_files_exist = True
        if os.path.exists(path=os.path.join(self.mesh_results_folder, "tmp_mesh_t.dat")):
            write_tmp_mas_file(mesh_type="t",
                               tmp_mesh_dir_name=os.path.join(self.mesh_results_folder, "tmp_mesh_t.dat"),
                               tmp_mas_dir_name=os.path.join(self.mesh_results_folder, "tmp_mas_t.dat"))
        else:
            all_tmp_mesh_files_exist = False

        if os.path.exists(path=os.path.join(self.mesh_results_folder, "tmp_mesh_p.dat")):
            write_tmp_mas_file(mesh_type="p",
                               tmp_mesh_dir_name=os.path.join(self.mesh_results_folder, "tmp_mesh_p.dat"),
                               tmp_mas_dir_name=os.path.join(self.mesh_results_folder, "tmp_mas_p.dat"))
        else:
            all_tmp_mesh_files_exist = False

        if os.path.exists(path=os.path.join(self.mesh_results_folder, "tmp_mesh_r.dat")):
            write_tmp_mas_file(mesh_type="r",
                               tmp_mesh_dir_name=os.path.join(self.mesh_results_folder, "tmp_mesh_r.dat"),
                               tmp_mas_dir_name=os.path.join(self.mesh_results_folder, "tmp_mas_r.dat"))
        else:
            all_tmp_mesh_files_exist = False

        return all_tmp_mesh_files_exist

    def write_header_file(self):
        """ write a mesh_header.dat file in folder. """
        write_mesh_header(tmp_mas_p_file=os.path.join(self.mesh_results_folder, "tmp_mas_p.dat"),
                          tmp_mas_t_file=os.path.join(self.mesh_results_folder, "tmp_mas_t.dat"),
                          tmp_mas_r_file=os.path.join(self.mesh_results_folder, "tmp_mas_r.dat"),
                          mesh_header_file_path=os.path.join(self.mesh_results_folder, "mesh_header.dat"))

    def write_hdf_files(self):
        """ write 1d mesh spacing hdf5 files and half mesh. """
        if os.path.exists(path=os.path.join(self.mesh_results_folder, "mesh_res_t.txt")) and \
                os.path.exists(path=os.path.join(self.mesh_results_folder, "mesh_res_p.txt")) and \
                os.path.exists(path=os.path.join(self.mesh_results_folder, "mesh_res_r.txt")):
            t, dt, t_ratio = read_tmp_file(name_of_tmp_file=os.path.join(self.mesh_results_folder, "mesh_res_t.txt"))
            p, dt, p_ratio = read_tmp_file(name_of_tmp_file=os.path.join(self.mesh_results_folder, "mesh_res_p.txt"))
            r, rt, r_ratio = read_tmp_file(name_of_tmp_file=os.path.join(self.mesh_results_folder, "mesh_res_r.txt"))
            write_1d_mesh_hdf(outdir=self.mesh_results_folder, mesh_name="1x", t=t, p=p, r=r)

    def get_all_tabs_info(self):
        """ copy all tabs information to the ui dictionary. this step is done before saving the json file. """
        # radial information.
        self.get_radial_rMin()
        self.get_radial_rMax()
        self.get_radial_fg_ratio()
        self.get_radial_bg_ratio()
        row_num = 1
        for row in self.radial_table_tb:
            self.read_radial_textbox_row(row, row_num)
            row_num += 1

        # theta and phi information.
        self.get_fg_ratio()
        self.get_bg_ratio()
        self.add_ds()

    def save_ui_session(self):
        """ save the current 3d mesh results in bin/mesh_header.dat and in a txt/json file in the ui folder. """
        # recalculate theta, phi, and radial mesh.
        try:
            self.theta(show_plot=False)
        except:
            print("Theta mesh is invalid.")
        try:
            self.phi(show_plot=False)
        except:
            print("Phi mesh is invalid.")
        try:
            self.radial(show_plot=False)
        except:
            print("Radial mesh is invalid. ")

        # write a tmp mas files:
        all_mas_files_exist = self.write_tmp_mas_files()

        # write_mesh_header file.
        if all_mas_files_exist:
            self.write_header_file()

        # write 1d hdf5 files with mesh spacing and half spacing.
        self.write_hdf_files()

        # save json ui_session in a txt/json file:
        root = tk.Tk()
        SaveDict(master=root, ui_session=self.ui_session, save_json_folder=self.mesh_results_folder)
        root.mainloop()

    def get_theta_results(self):
        try:
            self.theta()
        except:
            print("Theta mesh requirements are not valid. ")

    def get_phi_results(self):
        try:
            self.phi()
        except:
            print("Phi mesh requirements are not valid. ")

    def get_radial_results(self):
        try:
            self.radial()
        except:
            print("Radial mesh requirements are not valid. ")

    def theta(self, show_plot=True):
        """ generate a theta mesh. """
        self.get_bg_ratio()
        self.get_fg_ratio()
        self.add_ds()
        self.get_mesh_theta_1(show_plot)

    def get_mesh_theta_1(self, show_plot):
        """ include 4 steps of building a theta mesh. """
        if self.ui_session.BG_RATIO != 1.06:
            MeshSegment.DEFAULT_BG_REGION_RATIO = self.ui_session.BG_RATIO
        if self.ui_session.FG_RATIO != 1.03:
            MeshSegment.DEFAULT_FG_REGION_RATIO = self.ui_session.FG_RATIO
        mesh = Mesh(self.ui_session.lower_bnd_t, self.ui_session.upper_bnd_t, False)
        for box in self.ui_session.box_list:
            if 0 <= box.t0 <= np.pi and 0 <= box.t1 <= np.pi:
                mesh.insert_mesh_segment(MeshSegment(box.t0, box.t1, box.ds, var_ds_ratio=box.tp_var_ds))
        if mesh.is_valid():
            adj_mesh = mesh.resolve_mesh_segments().json_dict()
            legacy_mesh = mesh.build_legacy_mesh().json_dict()
            create_psi_mesh(adj_mesh, legacy_mesh, 't', self.mesh_results_folder, output_file_name="tmp_mesh_t.dat",
                            mesh_res_file_name="mesh_res_t.txt", show_plot=show_plot,
                            save_plot_path=self.mesh_results_folder,
                            save_plot=True, plot_file_name="theta_mesh_spacing.png")

    def phi(self, show_plot=True):
        """ generate a phi mesh."""
        self.get_bg_ratio()
        self.get_fg_ratio()
        self.add_ds()
        self.get_mesh_phi_1(show_plot)

    def get_mesh_phi_1(self, show_plot):
        """ include 4 steps of building a phi mesh. """
        if self.ui_session.BG_RATIO != 1.06:
            MeshSegment.DEFAULT_BG_REGION_RATIO = self.ui_session.BG_RATIO
        if self.ui_session.FG_RATIO != 1.03:
            MeshSegment.DEFAULT_FG_REGION_RATIO = self.ui_session.FG_RATIO

        mesh = Mesh(self.ui_session.lower_bnd_p, self.ui_session.upper_bnd_p, True)
        for box in self.ui_session.box_list:
            mesh.insert_mesh_segment(MeshSegment(box.p0, box.p1, box.ds, var_ds_ratio=box.tp_var_ds))

        if mesh.is_valid():
            adj_mesh = mesh.resolve_mesh_segments().json_dict()  # adjusted mesh as a dictionary
            legacy_mesh = mesh.build_legacy_mesh().json_dict()  # legacy mesh as a dictionary
            create_psi_mesh(adj_mesh, legacy_mesh, 'p', self.mesh_results_folder, output_file_name="tmp_mesh_p.dat",
                            mesh_res_file_name="mesh_res_p.txt", show_plot=show_plot, save_plot=True,
                            save_plot_path=self.mesh_results_folder, plot_file_name="phi_mesh_spacing.png")
        else:
            print("Phi input fails to build a mesh. Mesh is not valid.")

    def radial(self, show_plot=True):
        """ generate a radial mesh. """
        self.get_radial_bg_ratio()
        self.get_radial_fg_ratio()
        self.get_radial_rMin()
        self.get_radial_rMax()
        self.get_mesh_r_1(show_plot)

    def read_radial_textbox_row(self, row, row_num):
        """ check if radial input textbox content is valid."""
        try:
            s0 = float(row[0].get())
        except ValueError:
            s0 = None
        self.ui_session.box_list[row_num - 1].r0 = s0

        try:
            s1 = float(row[1].get())
        except ValueError:
            s1 = None
        self.ui_session.box_list[row_num - 1].r1 = s1

        try:
            ds = float(row[2].get())
        except ValueError:
            ds = None
        self.ui_session.box_list[row_num - 1].radial_ds = ds

        try:
            var_ds = float(row[3].get())
        except ValueError:
            var_ds = None
        self.ui_session.box_list[row_num - 1].radial_var_ds = var_ds

    def get_mesh_r_1(self, show_plot):
        """ include the 4 steps of building a mesh. """
        MeshSegment.DEFAULT_BG_REGION_RATIO = self.ui_session.RADIAL_BG_RATIO
        MeshSegment.DEFAULT_FG_REGION_RATIO = self.ui_session.RADIAL_FG_RATIO
        mesh = Mesh(self.ui_session.lower_bnd_r, self.ui_session.upper_bnd_r, False)
        row_num = 1
        for row in self.radial_table_tb:
            self.read_radial_textbox_row(row, row_num)
            s0 = self.ui_session.box_list[row_num - 1].r0
            s1 = self.ui_session.box_list[row_num - 1].r1
            ds = self.ui_session.box_list[row_num - 1].radial_ds
            var_ds = self.ui_session.box_list[row_num - 1].radial_var_ds
            if s0 is not None and s1 is not None:
                mesh.insert_mesh_segment(MeshSegment(s0, s1, ds=ds, var_ds_ratio=var_ds))
            row_num += 1
        if mesh.is_valid():
            adj_mesh = mesh.resolve_mesh_segments().json_dict()
            legacy_mesh = mesh.build_legacy_mesh().json_dict()
            create_psi_mesh(adj_mesh, legacy_mesh, 'r', self.mesh_results_folder, output_file_name="tmp_mesh_r.dat",
                            mesh_res_file_name="mesh_res_r.txt", show_plot=show_plot, save_plot=True,
                            save_plot_path=self.mesh_results_folder, plot_file_name="r_mesh_spacing.png")

    def add_ds(self):
        """ check if ds textbox content in theta/phi are valid."""
        i = 0
        for ds_box in self.ds_tab_array:
            try:
                ds_val = float(ds_box.get())
            except ValueError:
                ds_val = None
            self.ui_session.box_list[i].ds = ds_val
            i += 1

        i = 0
        for var_ds_box in self.var_ds_tab_array:
            try:
                var_ds_val = float(var_ds_box.get())
            except ValueError:
                var_ds_val = None
            self.ui_session.box_list[i].tp_var_ds = var_ds_val
            i += 1

    def get_bg_ratio(self):
        """ get the content in step 3 BG_RATIO textbox. (theta+phi)"""
        try:
            bg_ratio = float(self.bg_ratio.get())
        except ValueError:
            return
        self.ui_session.BG_RATIO = float(bg_ratio)

    def get_fg_ratio(self):
        """ get the content in step 3 FG_RATIO textbox. (theta+phi)"""
        try:
            fg_ratio = float(self.fg_ratio.get())
        except ValueError:
            return
        self.ui_session.FG_RATIO = float(fg_ratio)

    def get_radial_bg_ratio(self):
        """ get the content in RADIAL_BG_RATIO textbox. (radial)"""
        try:
            radial_bg_ratio = float(self.radial_bg_ratio.get())
        except ValueError:
            return
        self.ui_session.RADIAL_BG_RATIO = float(radial_bg_ratio)

    def get_radial_fg_ratio(self):
        """ get the content in RADIAL_FG_RATIO textbox. (radial)"""
        try:
            radial_fg_ratio = float(self.radial_fg_ratio.get())
        except ValueError:
            return
        self.ui_session.RADIAL_FG_RATIO = float(radial_fg_ratio)

    def get_radial_rMin(self):
        """ get the content in lower_bnd_r textbox. Radial mesh domain."""
        try:
            rMin = float(self.rMin.get())
        except ValueError:
            return
        self.ui_session.lower_bnd_r = float(rMin)

    def get_radial_rMax(self):
        """ get the content in upper_bnd_r textbox. Radial mesh domain."""
        try:
            rMax = float(self.rMax.get())
        except ValueError:
            return
        self.ui_session.upper_bnd_r = float(rMax)

    def redraw_submit(self):
        """submit the new coordinates of the redrawn box"""
        if self.box_num != 0:
            if len(self.ui_session.box_list) >= self.box_num:
                self.ax.add_patch(self.rect_data)
                rect_position = mpl.patches.Rectangle.get_bbox(self.rect_data)
                label_box = "Box # %s" % self.box_num
                self.ax.text(rect_position.x0, rect_position.y0, label_box)

                self.ui_session.box_list[self.box_num - 1].t0 = min(rect_position.y0, rect_position.y1)
                self.ui_session.box_list[self.box_num - 1].t1 = max(rect_position.y0, rect_position.y1)
                self.ui_session.box_list[self.box_num - 1].p0 = min(rect_position.x0, rect_position.x1)
                self.ui_session.box_list[self.box_num - 1].p1 = max(rect_position.x0, rect_position.x1)

                self.patch_list.insert(self.box_num - 1, self.rect_data)
                self.box_num = 0
                self.redraw_box_submit = False
        else:
            print("First click APPLY button to remove the box from image, then redraw the same box and click SUMBIT.")

    def redraw_box_apply(self):
        """ Remove the txt box number and patch of box. """
        if self.redraw_box_submit is False:
            try:
                box_num = self.redraw_drop_box_get_val()
            except ValueError:
                print("Input must be an integer")
                return

            if len(self.ui_session.box_list) < box_num or box_num <= 0:
                print("Index is out of range.")
            else:
                self.box_num = box_num
                i = 0
                while i < len(self.ax.texts):
                    if self.ax.texts[i].get_text() == "Box # " + str(box_num):  # remove txt of box number
                        self.ax.texts[i].remove()
                        break
                    i += 1

                current_box = self.patch_list[int(box_num) - 1]
                current_box.remove()
                self.patch_list.remove(current_box)
                self.redraw_box_submit = True
        else:
            print("Please redraw the box and then click submit.")

    def save_latest_box(self):
        """ Save the latest box drawn on the plot and append to ui_session box_list. """
        if self.box_num == 0:
            if self.rect_data == 0:
                print("First draw a rectangle on plot.")
                return
            else:
                self.ax.add_patch(self.rect_data)
                rect_position = mpl.patches.Rectangle.get_bbox(self.rect_data)
                i = (len(self.ui_session.box_list) + 1)
                label_box = "Box # %s" % i
                self.ax.text(rect_position.x0, rect_position.y0, label_box)
                self.ui_session.insert_mesh_box(i - 1, MeshBox(t0=min(rect_position.y0, rect_position.y1),
                                                               t1=max(rect_position.y0, rect_position.y1),
                                                               p0=min(rect_position.x0, rect_position.x1),
                                                               p1=max(rect_position.x0, rect_position.x1)))
                self.patch_list.append(self.rect_data)

                # add a ds text box for step #2
                tk.Label(self.leftFrame, text=label_box, relief=tk.RIDGE, bg='lightgray', bd=2). \
                    grid(row=8 + len(self.ui_session.box_list), column=0, sticky=tk.NSEW)
                self.ds_tab_array.append(tk.Entry(self.leftFrame, width=11, bg='white'))
                self.var_ds_tab_array.append(tk.Entry(self.leftFrame, width=11, bg='white'))
                self.ds_tab_array[-1].insert(tk.END, '')
                self.var_ds_tab_array[-1].insert(tk.END, '')
                self.ds_tab_array[-1].grid(row=8 + len(self.ui_session.box_list), column=1, sticky=tk.NSEW)
                self.var_ds_tab_array[-1].grid(row=8 + len(self.ui_session.box_list), column=2, sticky=tk.NSEW,
                                               columnspan=2)

                # add a radial textbox
                tk.Label(self.rightFrame, text=label_box, relief=tk.RIDGE, bg='lightgray', bd=2) \
                    .grid(row=10 + len(self.ui_session.box_list), column=0, sticky=tk.NSEW)
                cols = []
                i = len(self.ui_session.box_list) - 1
                for j in range(4):
                    e = tk.Entry(self.rightFrame, width=5)
                    e.insert(tk.END, '')
                    e.grid(row=10 + len(self.ui_session.box_list), column=j + 1, sticky=tk.NSEW)
                    cols.append(e)
                self.radial_table_tb.append(cols)

                # add option to dropbox in remove and redraw.
                self.update_delete_box_drop_box()
                self.update_redraw_box_drop_box()
                self.canvas.draw()

    def delete_box(self):
        """ Delete a box drawn on canvas. """
        try:
            box_id = int(self.drop_box_delete_box_get_val())  # check if input is valid
        except ValueError:
            print("The dropbox box number is invalid.")
            return

        if self.box_num == 0:
            if len(self.ui_session.box_list) < box_id or box_id <= 0:
                print("Index is out of range.")
            else:
                i = 0
                while i < len(self.ax.texts):
                    if self.ax.texts[i].get_text() == "Box # " + str(box_id):  # remove txt of box number
                        self.ax.texts[i].remove()
                        i += - 1
                    i += 1

                if int(box_id) < len(self.ui_session.box_list):  # adjust all text box numbers and ds boxes
                    idx = 0
                    for box in range(int(box_id) + 1, len(self.ui_session.box_list) + 1):
                        self.ds_tab_array[box - 2].delete(0, tk.END)
                        self.var_ds_tab_array[box - 2].delete(0, tk.END)
                        self.ds_tab_array[box - 2].insert(0, str(self.ds_tab_array[box - 1].get()))
                        self.var_ds_tab_array[box - 2].insert(0, str(self.var_ds_tab_array[box - 1].get()))
                        self.radial_table_tb[box - 2][0].delete(0, tk.END)
                        self.radial_table_tb[box - 2][1].delete(0, tk.END)
                        self.radial_table_tb[box - 2][2].delete(0, tk.END)
                        self.radial_table_tb[box - 2][3].delete(0, tk.END)
                        self.radial_table_tb[box - 2][0].insert(0, str(self.radial_table_tb[box - 1][0].get()))
                        self.radial_table_tb[box - 2][1].insert(0, str(self.radial_table_tb[box - 1][1].get()))
                        self.radial_table_tb[box - 2][2].insert(0, str(self.radial_table_tb[box - 1][2].get()))
                        self.radial_table_tb[box - 2][3].insert(0, str(self.radial_table_tb[box - 1][3].get()))
                        while idx < len(self.ax.texts):
                            if self.ax.texts[idx].get_text() == "Box # " + str(box):
                                self.ax.texts[idx].set_text("Box # " + str(box - 1))
                                break
                            idx += 1

                self.ui_session.box_list = np.delete(self.ui_session.box_list, int(box_id) - 1)
                current_box = self.patch_list[int(box_id) - 1]
                current_box.remove()
                self.patch_list.remove(current_box)

                # delete the lastest ds textbox and label
                tk.Label(self.leftFrame, text="             ").grid(row=9 + len(self.ui_session.box_list),
                                                                    column=0, sticky=tk.NSEW)
                self.ds_tab_array[-1].destroy()
                self.ds_tab_array.pop()
                self.var_ds_tab_array[-1].destroy()
                self.var_ds_tab_array.pop()

                # delete box label in radial table
                tk.Label(self.rightFrame, text="       ", bg='gray92', bd=2) \
                    .grid(row=11 + len(self.ui_session.box_list), column=0, sticky=tk.NSEW)
                for tab in self.radial_table_tb[-1]:
                    tab.destroy()
                self.radial_table_tb.pop()

                # delete the box option from dropbox in remove and redraw mesh.
                self.remove_menu_option_delete_dropbox()
                self.remove_menu_option_redraw_dropbox()
                self.canvas.draw()

    def box_chart(self):
        """ Window to edit the boxes coordinates manually. """
        root = tk.Tk()
        TableChart(root, self.ui_session, self.ax, self.patch_list)
        root.mainloop()

    def update_delete_box_drop_box(self):
        """ update delete box dropbox. """
        if len(self.ui_session.box_list) == 1:
            self.reset_delete_box_drop_box()
        choice = str(len(self.ui_session.box_list))
        self.del_tb_opt['menu'].add_command(label=choice, command=tk._setit(self.del_tb_variable, choice))

    def reset_delete_box_drop_box(self):
        """ reset the dropbox in delete box. """
        self.del_tb_variable.set('')
        self.del_tb_opt['menu'].delete(0, 'end')

    def drop_box_delete_box_get_val(self, *args):
        """ get the dropdown menu selected for delete box. """
        try:
            return int(self.del_tb_variable.get())
        except Exception:
            return 0

    def remove_menu_option_delete_dropbox(self):
        """ remove a box number from dropbox. """
        self.reset_delete_box_drop_box()
        for ii in range(1, len(self.ui_session.box_list) + 1):
            self.del_tb_opt['menu'].add_command(label=str(ii), command=tk._setit(self.del_tb_variable, str(ii)))

    def redraw_drop_box_get_val(self, *args):
        """ get the dropdown menu selected for redraw box. """
        try:
            return int(self.redraw_variable.get())
        except Exception:
            return 0

    def update_redraw_box_drop_box(self):
        """ update redraw box dropbox. """
        if len(self.ui_session.box_list) == 1:
            self.reset_redraw_box_drop_box()
        choice = str(len(self.ui_session.box_list))
        self.redraw_opt['menu'].add_command(label=choice, command=tk._setit(self.redraw_variable, choice))

    def reset_redraw_box_drop_box(self):
        """ reset the dropbox in redraw box. """
        self.redraw_variable.set('')
        self.redraw_opt['menu'].delete(0, 'end')

    def remove_menu_option_redraw_dropbox(self):
        """ once a box is deleted the options of boxes to redraw will change. """
        self.reset_redraw_box_drop_box()
        for ii in range(1, len(self.ui_session.box_list) + 1):
            self.redraw_opt['menu'].add_command(label=str(ii), command=tk._setit(self.redraw_variable, str(ii)))

    @staticmethod
    def set_colorbar(fig, im):
        """ add a colorbar to magnetogram plot. """
        position = fig.add_axes([0.94, 0.1, 0.01, 0.8])
        cb = fig.colorbar(im, cax=position, shrink=1.5, orientation="vertical")
        cb.ax.tick_params(labelsize=7)

    def setup_magnetogram_plot(self, fig, hdf_file_path):
        """ setup the plot in midframe. """
        self.set_axis_limits()

        cmin, im = self.read_hdf_file(hdf_file_path)

        # set a colorbar.
        self.set_colorbar(fig=fig, im=im)

        self.restrict_panning()
        return cmin

    def restrict_panning(self):
        """ restrict panning in x axis beyond limits. """
        RestrictPan(self.ax, x=lambda t: t <= 4 * np.pi)
        RestrictPan(self.ax, x=lambda t: t >= -2 * np.pi)

    def set_axis_limits(self):
        """set the x and y axis limits. """
        # Set the x,y limits and the aspect to equal length (this will flip the theta axis in the plot as desired)
        self.ax.set_xlim(0, np.pi * 2)
        self.ax.set_ylim(np.pi, 0.0)
        self.ax.set_aspect('equal')

    def read_hdf_file(self, hdf_file_path):
        """ read hdf magnetogram file. """
        # read the hdf file u
        hdfdata = plot_hdf(hdf_file_path)
        # Set the color scaling values
        cmin = self.initial_cmin_scale(hdfdata)
        self.im1 = self.plot_hdf_file(cmin, hdfdata)

        return cmin, self.im1

    @staticmethod
    def initial_cmin_scale(hdfdata):
        """ get initial hdf scaling. """
        cmin = np.min(hdfdata[2])
        cmax = np.max(hdfdata[2])

        if abs(cmin) > abs(cmax):
            cmax = cmax / 25
            cmin = -abs(cmax)
        if abs(cmax) > abs(cmin):
            cmin = cmin / 25

        return int(cmin)

    def plot_hdf_file(self, cmin, hdfdata):
        """ plot magnetogram."""
        self.im1 = self.ax.pcolorfast(hdfdata[0], hdfdata[1], hdfdata[2], cmap='RdBu', vmin=cmin, vmax=abs(cmin))
        self.im2 = self.ax.pcolorfast(2 * np.pi * np.ones(len(hdfdata[0])) + hdfdata[0], hdfdata[1], hdfdata[2],
                                      cmap='RdBu', vmin=cmin, vmax=abs(cmin))
        self.im3 = self.ax.pcolorfast(-2 * np.pi * np.ones(len(hdfdata[0])) + hdfdata[0], hdfdata[1], hdfdata[2],
                                      cmap='RdBu', vmin=cmin, vmax=abs(cmin))
        return self.im1

    def apply_new_scale(self):
        """ plot the magnetogram with new scaling. """
        new_value = self.read_new_cmax()
        if new_value == 0:
            print("cmin value is not valid.")
        else:
            self.im1.set_clim(-abs(new_value), abs(new_value))
            self.im2.set_clim(-abs(new_value), abs(new_value))
            self.im3.set_clim(-abs(new_value), abs(new_value))
            self.canvas.draw()

    def read_new_cmax(self):
        """ get the new cmin value from textbox"""
        try:
            return float(self.cmax_value.get())
        except Exception:
            return 0


if __name__ == "__main__":
    # Choose the files to read, point the base_path to where you put the MapPipeline folder
    pipeline_example_folder = '/Users/opalissan/Downloads/MagnetogramExamples'
    base_path = pipeline_example_folder + '/Oct2011'
    hdf_file_path1 = base_path + '/br_example_2011_10_01_raw.hdf'

    # ui dictionary
    my_ui_session = MeshInfo(lower_bnd_t=0, upper_bnd_t=np.pi, lower_bnd_p=0, upper_bnd_p=2 * np.pi,
                             lower_bnd_r=0, upper_bnd_r=2.5)

    app = MeshApp(hdf_file_path=hdf_file_path1, ui_session_data=my_ui_session, UseExistingInput=False,
                  save_mesh_file_path="/Users/opalissan/desktop/mesh_folder")



"""
Restrict the x limits of panning a matplotlib plot.
"""
import numpy as np


class RestrictPan:
    def __init__(self, ax, x=lambda x: True):
        self.res = [x]
        self.ax = ax
        self.limits = self.get_lim()
        self.ax.callbacks.connect('xlim_changed', lambda evt: self.limits_change(axis=0))

    def get_lim(self):
        return [self.ax.get_xlim(), self.ax.get_ylim()]

    def set_lim(self, axis, lim):
        if axis == 0:
            self.ax.set_xlim(lim)
        else:
            self.ax.set_ylim(lim)
        self.limits[axis] = self.get_lim()[axis]

    def limits_change(self, event=None, axis=0):
        cur_lim = np.array(self.get_lim()[axis])
        if self.limits[axis] != self.get_lim()[axis]:
            # avoid recursion
            if not np.all(self.res[axis](cur_lim)):
                # if limits are invalid, reset them to previous state
                self.set_lim(axis, self.limits[axis])
            else:
                # if limits are valid, update previous stored limits
                self.limits[axis] = self.get_lim()[axis]



""" Adjusted Mesh Segment

Building a mesh can be broken down into four steps.
Here, we will contribute to compute step 2, and 3.

In step 2: Adjust mesh segments.
    While running mesh/resolve_mesh_segments() if there is a need to add segments or delete segments then the class
    below is called.

In step 3: Build Legacy mesh.
    In the function below "get_legacy_segment()" it takes in a mesh segment with a specified domain,
    resolution, and ratio.
    Then computes its total number of points and the ratio between the begin and end ds (ds1/ds0).

The solution details are explained below:
    * Note: these equations are also in the mkdocs static site under "mesh equations".

    For each segment I want to know the number of points it will take me to go
    either a fixed distance at a certain cell-to-cell ratio, R, OR get to a
    fixed new ds at a certain R.

    The MAS mesh equation calculates the next segment ds as ds(i+1) = R*ds(i),
    where R is the cell to cell mesh expansion ratio.

    This looks like the equation for ds as a function of npts: ds(n) = ds(0)*R^n

    What I need is to determine the number of points for each segment that goes
    from s0 to s1, satisfying certain constrains.

    Currently I divide this into three cases:

    CONSTANT: simplest, R=1, n = (s1-s0)/ds0

    FIXED_LENGTH: Go a fixed distance at a specified ratio starting at a given
                  ds0. Integrate the ds(n) equation, get S = ds0*(R^n - 1)/ln(R)
                  Solve this for n, get n = ln[ ln(R)*S/ds0 +1]/ln(R)

    FIXED_DS: Specify the DS at both sides (ds0, and ds1) and attempt to find a
              Solution that does this in the fewest number of points, at a fixed R
            - Solve ds1 = ds0*R^n, which is n = ln(ds1/ds0)/ln(R).
              Then the length is S = ds0*(R^n - 1)/ln(R)
            - HOWEVER, the length is not constrained to be s1-s0 --> it will
              be shorter or longer.
              - shorter case is simple, add const mesh at larger ds to make up gap
              - longer case means you have to go back to previous or next seg
              ---> a solution may not be found if it crosses others.
                - solving this issue adds most of complexity of the code because
                  I have to add recursion and a lot of case checking.

"""

import json
import numpy
from .legacy_mesh_segment import LegacyMeshSegment

EPSILON = 1 / numpy.power(10, 10)


class MeshSegment:
    # Mesh points distance in background area. INF value indicates that the
    # mesh can be as coarse as possible.
    BG_REGION_DS = numpy.inf

    # Ratio of mesh with constant density.
    CONSTANT_DS_RATIO = 1.00

    # Default ratio of mesh with decreasing density (increasing ds) in area
    # of interest and background area respectively.
    DEFAULT_FG_REGION_RATIO = 1.03
    DEFAULT_BG_REGION_RATIO = 1.06

    def __init__(self, s0, s1, ds=None, var_ds_ratio=None):
        """
        float: s0 and s1
            Lower and upper boundaries respectively.
        float: ds
            Distance between two mesh points, which represents density.
        float: var_ds_ratio
            Ratio between two subsequent distances of mesh with decreasing
            density (increasing ds).
        """
        if ds is None:
            ds = MeshSegment.BG_REGION_DS

        if var_ds_ratio is None:
            var_ds_ratio = MeshSegment.DEFAULT_BG_REGION_RATIO \
                if ds == MeshSegment.BG_REGION_DS \
                else MeshSegment.DEFAULT_FG_REGION_RATIO

        self.s0 = numpy.float64(s0)
        self.s1 = numpy.float64(s1)
        self.ds0 = numpy.float64(ds)
        self.ds1 = numpy.float64(ds)
        self.var_ds_ratio = numpy.float64(var_ds_ratio)

    def __str__(self):
        return json.dumps(
            self.json_dict(), indent=2, default=lambda o: o.json_dict())

    def json_dict(self):
        return {
            's0': self.s0,
            's1': self.s1,
            'ds0': self.ds0,
            'ds1': self.ds1,
            'var_ds_ratio': self.var_ds_ratio,
            'ratio': self.get_ratio()
        }

    @staticmethod
    def validate_bnd(s0, s1):
        return s0 < s1

    @staticmethod
    def validate_ds(ds0, ds1=None):
        if ds0 <= 0.0:
            return False

        if ds1 is not None and ds1 <= 0.0:
            return False

        return True

    def set_ds(self, ds0, ds1=None):
        """set the mesh segments ds0 and ds1 (begin and end resolution respectively)."""
        if ds1 is None:
            ds1 = ds0

        self.ds0 = numpy.float64(ds0)
        self.ds1 = numpy.float64(ds1)

    @staticmethod
    def validate_var_ds_ratio(var_ds_ratio):
        return var_ds_ratio > 0.0 and var_ds_ratio != MeshSegment.CONSTANT_DS_RATIO

    def set_var_ds_ratio(self, var_ds_ratio):
        self.var_ds_ratio = numpy.float64(var_ds_ratio)

    def segment_is_valid(self):
        return self.validate_bnd(self.s0, self.s1) \
               and self.validate_ds(self.ds0, self.ds1) \
               and self.validate_var_ds_ratio(self.var_ds_ratio)

    def get_ratio(self):
        """
        Returns the effective ratio based on the ds values on both end points.
        A mesh with constant density has a ratio of 1.0. Otherwise, the ratio
        will be according to the user specification.
        """
        if self.has_increasing_ds():
            return self.var_ds_ratio if self.var_ds_ratio > 1.0 \
                else 1.0 / self.var_ds_ratio
        elif self.has_decreasing_ds():
            return self.var_ds_ratio if self.var_ds_ratio < 1.0 \
                else 1.0 / self.var_ds_ratio
        else:
            return MeshSegment.CONSTANT_DS_RATIO

    def is_fixed_length(self):
        """
        Returns whether this segment has an unspecified ds value on one of the
        end points. This represents a background area where the mesh can be as
        coarse as possible.
        """
        return (self.ds0 == MeshSegment.BG_REGION_DS \
                or self.ds1 == MeshSegment.BG_REGION_DS) \
               and self.ds0 != self.ds1

    def is_fixed_ds(self):
        """
        Returns whether this segment has specific and different ds values on
        both end points. This represents an area of interest where the mesh
        is either decreasing or increasing in density.
        """
        return self.ds0 != MeshSegment.BG_REGION_DS \
               and self.ds1 != MeshSegment.BG_REGION_DS \
               and self.ds0 != self.ds1

    def is_constant(self):
        """
        Returns whether this segment has one specific ds value on both end
        points. This represents an area of interest where the mesh has constant
        density i.e., distributed evenly across the region.
        """
        return self.ds0 != MeshSegment.BG_REGION_DS and self.ds0 == self.ds1

    def has_increasing_ds(self):
        """
        Returns whether the mesh in this segment has increasing ds values i.e.,
        decreasing density.
        """
        return self.ds0 < self.ds1

    def has_decreasing_ds(self):
        """
        Returns whether the mesh in this segment has decreasing ds values i.e.,
        increasing density.
        """
        return self.ds0 > self.ds1

    def get_ds0_using_num_points(self, num_points):
        """
        Returns ds0 using the segments number of points and ratio. In this case usually, ds0 = inf.
        """
        ratio = 1 / self.get_ratio()
        return self.ds1 * numpy.power(ratio, num_points)

    def get_ds1_using_num_points(self, num_points):
        """
        Returns ds1 using the segments number of points and ratio. In this case usually, ds1 = inf.
        """
        return self.ds0 * numpy.power(self.get_ratio(), num_points)

    def get_length(self):
        """
        Returns the segment length.
        """
        return self.s1 - self.s0

    def get_legacy_ratio(self):
        """
        Returns the segment ratio between mesh distances at the two end points.
        """
        if self.is_constant():
            return 1.0
        elif self.is_fixed_ds():
            return self._get_legacy_ratio_fixed_ds_mesh()
        elif self.is_fixed_length():
            return self._get_legacy_ratio_fixed_length_mesh()

        raise Exception('Unknown mesh segment type.')

    def get_num_points(self):
        """
        Returns the minimum number of mesh points this segment would need to
        meet its constraints.
        """
        if self.is_constant():
            return self._get_num_points_constant_mesh()
        elif self.is_fixed_ds():
            return self._get_num_points_fixed_ds_mesh()
        elif self.is_fixed_length():
            return self.get_num_points_fixed_length_mesh()

        raise Exception('Unknown mesh segment type.')

    def get_legacy_mesh_segments(self, lower_bnd, upper_bnd):
        """
        Important function to compute legacy mesh segment.
        Here we compute the 3 mesh segment types: constant/fixed_ds/fixed_length.
        It handles the overshooting fixed_ds cases by adding a small constant segment to fill in the gap.
        """
        legacy_mesh_segment_list = numpy.array([])

        if self.is_constant() or self.is_fixed_length():
            legacy_segment = LegacyMeshSegment(
                self.s0, self.s1, self.get_legacy_ratio(),
                self.get_num_points())
            legacy_mesh_segment_list = numpy.append(
                legacy_mesh_segment_list, legacy_segment)
            return legacy_mesh_segment_list

        segment_num_points = self.get_num_points()
        legacy_ratio = self.get_legacy_ratio()
        segment_length = self.get_length()
        length_var_ds = self._get_length_using_num_points(segment_num_points)

        if length_var_ds < segment_length - EPSILON:
            if self.has_increasing_ds():
                length_cst_ds = self.s0 + length_var_ds
                new_cst_segment = MeshSegment(
                    length_cst_ds, self.s1, self.ds1)
                legacy_segment = LegacyMeshSegment(
                    self.s0, new_cst_segment.s0, legacy_ratio,
                    segment_num_points)
                legacy_new_cst_segment = LegacyMeshSegment(
                    new_cst_segment.s0, new_cst_segment.s1,
                    new_cst_segment.get_legacy_ratio(),
                    new_cst_segment.get_num_points())
                legacy_mesh_segment_list = numpy.append(
                    legacy_mesh_segment_list, legacy_segment)
                legacy_mesh_segment_list = numpy.append(
                    legacy_mesh_segment_list, legacy_new_cst_segment)
            else:
                length_cst_ds = self.s1 - length_var_ds
                new_cst_segment = MeshSegment(
                    self.s0, length_cst_ds, self.ds0)
                legacy_segment = LegacyMeshSegment(
                    new_cst_segment.s1, self.s1, legacy_ratio,
                    segment_num_points)
                legacy_new_cst_segment = LegacyMeshSegment(
                    new_cst_segment.s0, new_cst_segment.s1,
                    new_cst_segment.get_legacy_ratio(),
                    new_cst_segment.get_num_points())
                legacy_mesh_segment_list = numpy.append(
                    legacy_mesh_segment_list, legacy_new_cst_segment)
                legacy_mesh_segment_list = numpy.append(
                    legacy_mesh_segment_list, legacy_segment)
        elif length_var_ds > segment_length + EPSILON:
            if self.has_increasing_ds():
                if self.s1 == upper_bnd:
                    self.ds1 = numpy.inf
                    legacy_segment = LegacyMeshSegment(
                        self.s0, self.s1, legacy_ratio,
                        segment_num_points)
                else:
                    legacy_segment = LegacyMeshSegment(
                        self.s0, self.s0 + length_var_ds, legacy_ratio,
                        segment_num_points)
            else:
                if self.s0 == lower_bnd:
                    self.ds0 = numpy.inf
                    legacy_segment = LegacyMeshSegment(
                        self.s0, self.s1, legacy_ratio,
                        segment_num_points)
                else:
                    legacy_segment = LegacyMeshSegment(
                        self.s1 - length_var_ds, self.s1, legacy_ratio,
                        segment_num_points)

            legacy_mesh_segment_list = numpy.append(
                legacy_mesh_segment_list, legacy_segment)
        else:
            legacy_segment = LegacyMeshSegment(
                self.s0, self.s1, legacy_ratio,
                segment_num_points)
            legacy_mesh_segment_list = numpy.append(
                legacy_mesh_segment_list, legacy_segment)

        return legacy_mesh_segment_list

    def _get_legacy_ratio_fixed_ds_mesh(self):
        """return the segment ratio when segment is of fixed_ds type."""
        return self.ds1 / self.ds0

    def _get_legacy_ratio_fixed_length_mesh(self):
        """return the segment ratio when segment is of fixed_length type."""
        return numpy.power(self.get_ratio(), self.get_num_points())

    def _get_length_using_num_points(self, num_points):
        """
        By knowing the segment's total number of points and ratio we can compute its domain length.
        S = ds0*(R^n - 1)/ln(R)
        """
        if self.is_constant():
            return self.ds0 * num_points
        else:
            if self.has_decreasing_ds():
                ds = self.ds1
                ratio = 1.0 / self.get_ratio()
            else:
                ds = self.ds0
                ratio = self.get_ratio()

        return ds * (numpy.power(ratio, num_points) - 1) / numpy.log(ratio)

    def _get_num_points_constant_mesh(self):
        """Compute the constant mesh segment total number of points."""
        return int(numpy.ceil(self.get_length() / self.ds0))

    def _get_num_points_fixed_ds_mesh(self):
        """Compute the fixed_ds mesh segment total number of points."""
        return int(numpy.ceil(
            numpy.log(self.ds1 / self.ds0) / numpy.log(self.get_ratio())))

    def get_num_points_fixed_length_mesh(self):
        """Compute the fixed_length mesh segment total number of points."""
        if self.has_decreasing_ds():
            ds = self.ds1
            ratio = 1.0 / self.get_ratio()
        else:
            ds = self.ds0
            ratio = self.get_ratio()
        ln_ratio = numpy.log(ratio)

        return int(numpy.ceil(numpy.log(ln_ratio * self.get_length() / ds + 1) / ln_ratio))
